// ---------------------------------------------------------------------------
// NNFConverter.java
// Conversion of LTL formulas into negation normal form.
// $Id: NNFConverter.java,v 1.1 2022/08/24 12:47:14 schreine Exp $
//
// Author: Ágoston Sütő <agoston.suetoe@jku.at>
// Copyright (C) 2021-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package riscal.ltl.automaton;

import riscal.ltl.LTLFormula;
import riscal.ltl.LTLProcessor;

import static riscal.ltl.LTLFormula.LTLOperator.*;

/**
 * The automaton creation algorithm used in this implementation
 * {@link AutomatonCreator} requires the formula to be in negation
 * normal form. This class transforms a given LTL formula into
 * NNF by pushing the `not` operators to in front of the state
 * and action formulas using de Morgan's rules: `and` and `or`
 * are duals, `until` and `release` are duals, `next time` is
 * its own dual.
 */
public class NNFConverter {

	private static final NNFConverterImpl converter = new NNFConverterImpl();

	public static LTLFormula convertToNNF(LTLFormula LTL) {
		return LTL.process(converter, false);
	}

	private static class NNFConverterImpl implements LTLProcessor<LTLFormula, Boolean> {
		@Override
		public LTLFormula process(LTLFormula.BooleanNode LTL, Boolean negated) {
			return (LTL.value ^ negated) ? LTLFormula.BooleanNode.TRUE : LTLFormula.BooleanNode.FALSE;
		}

		@Override
		public LTLFormula process(LTLFormula.NegationNode LTL, Boolean negated) {
			return LTL.child.process(this, !negated);
		}

		@Override
		public LTLFormula process(LTLFormula.NextNode LTL, Boolean negated) {
			return new LTLFormula.NextNode(LTL.child.process(this, negated));
		}

		@Override
		public LTLFormula process(LTLFormula.OperatorNode LTL, Boolean negated) {
			LTLFormula.LTLOperator operator = negated ? getDual(LTL.operator) : LTL.operator;
			LTLFormula left = LTL.left.process(this, negated);
			LTLFormula right = LTL.right.process(this, negated);

			return new LTLFormula.OperatorNode(operator, left, right);
		}

		@Override
		public LTLFormula process(LTLFormula.StateNode LTL, Boolean negated) {
			return negated ? new LTLFormula.NegationNode(LTL) : LTL;
		}

		@Override
		public LTLFormula process(LTLFormula.ActionNode LTL, Boolean negated) {
			return negated ? new LTLFormula.NegationNode(LTL) : LTL;
		}

		private static LTLFormula.LTLOperator getDual(LTLFormula.LTLOperator operator) {
			switch (operator) {
				case AND:
					return OR;
				case OR:
					return AND;
				case UNTIL:
					return RELEASE;
				case RELEASE:
					return UNTIL;
				default:
					throw new IllegalArgumentException();
			}
		}
	}
}
// ----------------------------------------------------------------------------
// end of file
// ----------------------------------------------------------------------------

