// ---------------------------------------------------------------------------
// CheckingState.java
// A state in the product automaton of the LTL model checker.
// $Id: CheckingState.java,v 1.2 2023/05/09 09:23:03 schreine Exp $
//
// Author: Ágoston Sütő <agoston.suetoe@jku.at>
// Copyright (C) 2021-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package riscal.ltl.search;

import riscal.ltl.LTLChecker;
import riscal.ltl.automaton.AutomatonNode;
import riscal.semantics.Systems;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Objects;

/**
 * A state in the product automaton of the system and the formula.
 * Used for the non-emptyness check. CheckingStates are unique,
 * meaning checkingState1.equals(checkingState2) implies that
 * checkingState1 == checkingState2. This uniqueness is ensured
 * by the `checkingStateCache` in {@link LTLChecker}.
 */
public class CheckingState {

    private final LTLChecker ltlChecker;
    final AutomatonNode automatonNode;
    final Systems.StateClass stateNode;

    private List<CheckingState> successors;
    private int processedSuccessors;

    int dfsnum;
    boolean current;

    public CheckingState(LTLChecker ltlChecker, AutomatonNode automatonNode, Systems.StateClass stateNode) {
        this.ltlChecker = ltlChecker;
        this.automatonNode = automatonNode;
        this.stateNode = stateNode;
    }

    public CheckingState getFirstSuccessor() {
        processedSuccessors = 0;
        return getNextSuccessor();
    }

    public CheckingState getNextSuccessor() {
        return processedSuccessors < getSuccessors().size() ? getSuccessors().get(processedSuccessors++) : null;
    }

    public List<CheckingState> getSuccessors() {
        if (successors == null) {
            List<Systems.StateClass> systemStates = ltlChecker.getSuccessors(stateNode);
            if (systemStates.isEmpty()) {
                ltlChecker.emitDeadlockWarning(stateNode);
                systemStates = Collections.singletonList(stateNode);
            }

            successors = new ArrayList<>();

            for (Systems.StateClass systemState : systemStates) {
                for (AutomatonNode automatonState : automatonNode.children) {
                    if (!automatonState.isCompatibleWith(ltlChecker, systemState)) {
                        continue;
                    }

                    successors.add(
                            ltlChecker.checkingStateCache.computeIfAbsent(
                                    new CheckingState(ltlChecker, automatonState, systemState),
                                    x -> x
                            )
                    );
                }
            }
        }

        return successors;
    }

    public boolean isUnprocessed() {
        return successors == null;
    }

    private boolean stateNodesEqual(Systems.StateClass a, Systems.StateClass b) {
        return Objects.equals(a, b)
                && a.label.action == b.label.action
                && a.label.instance == b.label.instance
                && Arrays.equals(a.label.args, b.label.args);
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;
        CheckingState state = (CheckingState) o;
        return automatonNode.equals(state.automatonNode) && stateNodesEqual(stateNode, state.stateNode);
    }

    @Override
    public int hashCode() {
        String transitionName = stateNode.label != null ? stateNode.label.action.ident.string : null;
        return Objects.hash(automatonNode, stateNode, transitionName);
    }
}
// ----------------------------------------------------------------------------
// end of file
// ----------------------------------------------------------------------------