// ---------------------------------------------------------------------------
// LTLError.java
// A violating system execution of the LTL model checker.
// $Id: CounterExample.java,v 1.2 2023/05/09 09:23:03 schreine Exp $
//
// Author: Ágoston Sütő <agoston.suetoe@jku.at>
// Copyright (C) 2021-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package riscal.ltl.search;

import java.util.Deque;
import java.util.List;

/**
 * Class representing a system execution which violates
 * a given temporal property.
 * A counter-example theoretically consists of two parts,
 * one being the path to the cycle and the other the
 * cycle itself. Here for technical reasons the prefix is
 * split into two parts.
 */
public class CounterExample {
    private final Deque<CheckingState> pathToScc;
    private final List<CheckingState> pathToCycle;
    private final List<CheckingState> cycle;

    CounterExample(Deque<CheckingState> pathToScc, List<CheckingState> pathToCycle, List<CheckingState> cycle) {
        this.pathToScc = pathToScc;
        this.pathToCycle = pathToCycle;
        this.cycle = cycle;

        // remove the dummy initial node (the "last" is actually the first for us)
        pathToScc.removeLast();
        if (!pathToCycle.isEmpty()) {
            // pathToCycle ends in the first state of the loop, so we drop it
            pathToCycle.remove(pathToCycle.size() - 1);
        }
        if (cycle != null) {
            // cycle is of the form a -> b -> ... -> a, and we don't want to display a twice
            cycle.remove(0);
        }
    }

    public String toString(String[] variables) {
        StringBuilder stringBuilder = new StringBuilder();

        pathToScc.descendingIterator().forEachRemaining(prefixState ->
                printState(stringBuilder, variables, prefixState));

        for (CheckingState prefixState : pathToCycle) {
            printState(stringBuilder, variables, prefixState);
        }

        if (cycle != null) {
            stringBuilder.append("> Loop start\n");

            for (CheckingState loopState : cycle) {
                stringBuilder.append('\t');
                printState(stringBuilder, variables, loopState);
            }

            stringBuilder.append("> Loop end");
        }

        return stringBuilder.toString();
    }

    private void printState(StringBuilder stringBuilder, String[] variables, CheckingState state) {
        stringBuilder
                .append("Action: ")
                .append(state.stateNode.label == null ? "init()" : state.stateNode.label)
                .append(" values: ")
                .append(state.stateNode.toString(variables))
                .append('\n');
    }
}
// ----------------------------------------------------------------------------
// end of file
// ----------------------------------------------------------------------------
