// ---------------------------------------------------------------------------
// RISCTPControl.java
// The control of the RISCTP theorem proving interface.
// $Id: RISCTPControl.java,v 1.16 2023/03/08 07:42:37 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.jku.at>
// Copyright (C) 2022-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, https://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package riscal.risctp;

import java.io.*;
import java.util.*;
import java.util.function.*;

import java.util.concurrent.*;

import riscal.*;
import riscal.syntax.AST.*;
import riscal.syntax.AST.Declaration.*;

public class RISCTPControl
{
  /****************************************************************************
   * Prove theorem in context of declarations with the possibility of
   * being stopped by pressing the "stop" button.
   * @param theorem the theorem.
   * @param decls the preceding declarations.
   * @param writer the output medium for the prover.
   * @return true if and only if the theorem could be proved.
   ***************************************************************************/
  public static boolean proveStoppable(TheoremParamDefinition theorem, List<Declaration> decls,
    PrintWriter writer)
  {
    boolean valid[] = new boolean[] { false };
    executeStoppable(new Runnable() {
      public void run()
      {
        PrintWriter writer0 = Main.getOutput();
        writer0.println("Proving theorem " + theorem.ident + "...");
        long time = System.currentTimeMillis();
        valid[0] = prove(theorem, decls, writer);
        time = System.currentTimeMillis()-time;
        if (valid[0])
        {
          writer0.println("SUCCESS: theorem was proved (" + time + " ms, see 'Print Prover Output').");
          int mode = RISCTPOptions.getMode();
          if (mode == 0) writer0.println("WARNING: only type-checking theorems were proved (see TP Proof Mode).");
          else if (mode == 1) writer0.println("WARNING: no type-checking theorems were proved (see TP Proof Mode).");
        }
        else
          writer0.println("ERROR: theorem was NOT proved (" + time + " ms, see 'Print Prover Output').");
      }});
    return valid[0];
  }
 
  /****************************************************************************
   * Perform an activity with the possibility of being stopped 
   * by pressing the "stop" button.
   * @param runnable the activity.
   ***************************************************************************/
  @SuppressWarnings("deprecation")
  private static void executeStoppable(Runnable runnable)
  {
    Main.resetStopped();
    Thread thread = new Thread(runnable);
    thread.start();
    try
    {
      while (true)
      {
        Thread.sleep(100); 
        if (!thread.isAlive()) break;
        // throws exception if "stop" button was pressed
        Main.checkStopped(); 
      }
    }
    catch(Exception e) { }
    if (Main.wasStopped()) { Main.getOutput().println("ERROR: execution was stopped."); }
    if (!thread.isAlive()) return;
    thread.interrupt();
    try { Thread.sleep(500); } catch(InterruptedException e0) { }
    if (thread.isAlive()) 
    {
      Main.getOutput().println("Thread could not be gracefully interrupted; it is forcefully stopped.");
      thread.stop();
    }      
  }
  
  // the list of tasks to be processed
  private static List<Runnable> tasks;
  
  /****************************************************************************
   * Initialize a pool of proving tasks
   ***************************************************************************/
  public static void initTasks()
  {
    tasks = new ArrayList<Runnable>();
  }
  
  /***************************************************************************
   * Add proving task to pool.
   * @param theorem the theorem to be proved.
   * @param decls the preceding declarations.
   * @param writer the output medium for the prover.
   * @param callback the function to be called with the task result.
   **************************************************************************/
  public static void addTask(TheoremParamDefinition theorem, 
    List<Declaration> decls, PrintWriter writer, Consumer<Boolean> callback)
  {
    tasks.add(new Runnable() {
      public void run()
      {
        boolean valid = prove(theorem, decls, writer);
        callback.accept(valid);
      }
    });
  }
  
  /***************************************************************************
   * Prove all tasks in pool
   **************************************************************************/
  public static void executeTasks()
  {
    PrintWriter writer = Main.getOutput();
    Main.resetStopped();
    int T = Main.getThreads();
    writer.println("Parallel execution with " + T + " threads (no output is shown)...");
    ExecutorService pool = Executors.newFixedThreadPool(T);
    long time = System.currentTimeMillis();
    for (Runnable task : tasks) pool.execute(task);
    tasks = null;
    pool.shutdown();
    try
    {
      while (true) 
      { 
        Thread.sleep(100); 
        if (pool.isTerminated()) break;
        // throws exception if "stop" button was pressed
        Main.checkStopped();
      }
    }
    catch(Exception e) { }
    time = System.currentTimeMillis()-time;
    if (Main.wasStopped()) 
     writer.println("ERROR: execution was stopped.");
    else
     writer.println("Execution completed (" + time + " ms, see 'Print Prover Output').");
    if (pool.isTerminated()) return;
    pool.shutdownNow();
  }
  
  /****************************************************************************
   * Prove theorem in context of declarations.
   * @param theorem the theorem.
   * @param decls the preceding declarations.
   * @param callers the caller map of t
   * @param write the output medium for the prover.
   * @return true if and only if the theorem could be proved.
   ***************************************************************************/
  public static boolean prove(TheoremParamDefinition theorem, List<Declaration> decls,
    PrintWriter writer)
  {
    writer.println(risctp.Main.COPYRIGHT);
    writer.println("Proving theorem " + theorem.ident + "...");
    int mode = RISCTPOptions.getMode();
    
    // configure generic settings of theorem proving interface 
    risctp.Main main = new risctp.Main();
    main.setInputOutput(null, writer);
    // main.setOptionQuiet(Main.getSilent());
    main.setMode(mode);
    // theorem names may be overloaded
    riscal.types.Environment.Symbol.FunctionSymbol fsymbol =
        (riscal.types.Environment.Symbol.FunctionSymbol)theorem.ident.getSymbol();
    main.addTheorem("'" + riscal.Main.shortName(fsymbol) + "'");
    // main.setOptionPrintAll(true);
    
    // translation to RISCTP proof problem
    List<Declaration> decls0 = new ArrayList<Declaration>(decls);
    decls0.add(theorem);
    risctp.problem.ProofProblem problem = RISCTP.translate(decls0);
    if (problem == null) return false;
    
    // start timing for the proof result to be determined
    long time = System.currentTimeMillis();
    boolean okay = false;
    
    // prove by method SMT
    if (!okay && RISCTPOptions.getMethodSMT())
    {
      main.setMethod(risctp.Main.Method.SMT);
      /* if (!Main.getSilent()) */ main.setOptionPrintAP(true);
      /* if (!Main.getSilent()) */ main.setOptionPrintSMT(true);
      Integer timeout = RISCTPOptions.getTimeout();
      risctp.smt.Solver.setTimeout(timeout == null ? 0 : timeout);
      if (RISCTPOptions.getSolverZ3())
      {
        risctp.smt.Solver.setPath(RISCTPOptions.getExecPathZ3());
        risctp.smt.Solver.configure(risctp.smt.Solver.SolverType.Z3);
      }
      else if (RISCTPOptions.getSolverCVC5())
      {
        risctp.smt.Solver.setPath(RISCTPOptions.getExecPathCVC5());
        risctp.smt.Solver.configure(risctp.smt.Solver.SolverType.CVC5);
      }
      else if (RISCTPOptions.getSolverVampire())
      {
        risctp.smt.Solver.setPath(RISCTPOptions.getExecPathVampire());
        risctp.smt.Solver.configure(risctp.smt.Solver.SolverType.VAMPIRE);
      }
      okay = main.prove(problem);
    }
    
    // if not yet proved, prove by method FOL
    if (!okay && RISCTPOptions.getMethodFOL())
    {
      main.setMethod(risctp.Main.Method.ME);
      /* if (!Main.getSilent()) */ main.setOptionPrintPR(true);
      okay = main.prove(problem);
    }
    
    // if no method was selected, print pruned translation result
    if (!RISCTPOptions.getMethodSMT() && !RISCTPOptions.getMethodFOL())
    {
      main.setMethod(risctp.Main.Method.NONE);
      /* if (!Main.getSilent()) */ main.setOptionPrintPR(true);
      okay = main.prove(problem);
    }
    
    // show outcome
    time = System.currentTimeMillis()-time;
    if (okay)
    {
      writer.println("SUCCESS: theorem was proved (" + time + " ms).");
      if (mode == 0) writer.println("WARNING: only type-checking theorems were proved (see TP Proof Mode).");
      else if (mode == 1) writer.println("WARNING: no type-checking theorems were proved (see TP Proof Mode).");
    }
    else
      writer.println("ERROR: theorem was NOT proved (" + time + " ms).");
    return okay;
  }
}
// ---------------------------------------------------------------------------
// end of file
// ---------------------------------------------------------------------------