// ---------------------------------------------------------------------------
// FormulaTree.java
// The visualization of formula evaluations.
// $Id: FormulaTree.java,v 1.20 2018/12/10 08:37:07 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.jku.at>
// Copyright (C) 2018-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package riscal.semantics;

import riscal.syntax.*;
import riscal.types.Environment.Symbol.*;
import riscal.semantics.Types.*;
import java.util.*;

public class FormulaTree
{
  // -------------------------------------------------------------------------
  // a tree node
  // -------------------------------------------------------------------------
  public static class Node
  {
    public final Node parent;     // may be null
    public final int cnumber;     // the number of the child
    private List<Node> children;  // the children nodes
    private int cvalues;          // the children with result values
    public final Context context; // the visible variables
    public final String label;    // the label of the node
    private AST phrase;           // the phrase attached to the node
    public final Cut cut;         // the permission to cut branches
    private Boolean value;        // the result value (potentially null)
    private Node(Node parent, Context context, String label, AST phrase, Cut cut)
    {
      this.parent = parent;
      if (parent == null) 
        cnumber = 0;
      else
      {
        cnumber = parent.children.size();
        parent.children.add(this);
      }
      this.children = new ArrayList<Node>();
      this.cvalues = 0;
      this.context = context == null ? null : new Context(context);
      this.label = label;
      this.phrase = phrase;
      this.cut = cut;
      value = null;
    }
    private void next(AST phrase, Context context, boolean value)
    {
      cvalues++;
      if (cvalues == children.size()) 
      {  
        Node child = children.get(cvalues-1);
        if (child.phrase == null) child.phrase = phrase;
        return;
      }
      Node child = new Node.Parent(this, context, phrase);
      child.close(value);
    }
    protected void close(boolean value)
    {
      this.value = value;
      if (cut == Cut.None) return;
      if (cut == Cut.Or  && value != true)  return;
      if (cut == Cut.And && value != false) return;
      int n = children.size();
      if (n == 0) return;
      if (cut == Cut.Implies)
      {
        Node firstNode = children.get(0);
        if (!firstNode.getValue()) 
        {
          children = new ArrayList<Node>();
          children.add(firstNode);
          return;
        }
        Node lastNode = children.get(n-1);
        if (lastNode.getValue()) 
        {
          children = new ArrayList<Node>();
          children.add(lastNode);
        }
      }
      else
      {
        Node lastNode = children.get(n-1);
        children = new ArrayList<Node>();
        children.add(lastNode);
      }
    }
    protected void print(int level)
    {
      for (int i=0; i<level; i++) System.out.print("  ");
      System.out.println(level + "/" + cnumber + ":" 
          + label + "(" + cut + ")");
      for (Node child : children)
        child.print(level+1);
    }
    public List<Node> getChildren() { return children; }
    public AST getPhrase() { return phrase; }
    public boolean hasValue() { return value != null; }
    public boolean getValue() { return value; }
    
    // some special nodes
    public static class Parent extends Node
    {
      private Parent(Node parent, Context context, AST phrase)
      {
        super(parent, context, "(parent)", phrase, Cut.None);
      }
    }
    public static class Application extends Node
    {
      private Value result;
      private Application(Node parent, Context context, String label, AST phrase, Cut cut)
      {
        super(parent, context, label, phrase, cut);
      }
      protected void close(boolean value) 
      { 
        this.result = new Value.Bool(value);
        super.close(value);
      }
      protected void close(Value value) 
      { 
        if (!(value instanceof Value.Bool))
          close(false);
        else
          close((((Value.Bool)value).getValue()));
      }
      public void setResult(Value result) { this.result = result; }
      public Value getResult() { return result; }
    }
    public static class Arguments extends Node
    {
      public final ValueSymbol[] params;
      public final Value[] values;
      private Value result;
      private Arguments(Node parent, Context context, String label, AST phrase, Cut cut,
        ValueSymbol[] params, Value[] values)
      {
        super(parent, context, label, phrase, cut);
        this.params = params;
        this.values = values;
      }
      public void setResult(Value result) { this.result = result; }
      public Value getResult() { return result; }
    }
  }
  
  // the root node (may be null, then no visualization)
  private static Node.Application root = null;
  
  // the current node (null)
  private static Node current = null;
  
  // the cut values (no cut, cut if some result is true, c
  // cut if some result is false, cut if first result is false)
  public static enum Cut { None, Or, And, Implies }
  
  /***************************************************************************
   * Return true if and only if there is a tree.
   **************************************************************************/
  public static boolean exists()
  {
    return root != null;
  }
  
  /***************************************************************************
   * Create new formula tree (deleting any previous one)
   **************************************************************************/
  public static void create(AST phrase)
  {
    current = root = new Node.Application(null, null, "(root)", phrase, Cut.And);
  }
  
  // the saved current node and root
  private static Node old = null;
  private static Node.Application oldroot = null;
  
  /***************************************************************************
   * Temporarily disable tree extension.
   **************************************************************************/
  public static void disable()
  {
    old = current;
    oldroot = root;
    current = null;
    root = null;
  }
  
  /***************************************************************************
   * Re-enable tree extension.
   **************************************************************************/
  public static void enable()
  {
    current = old;
    root = oldroot;
    old = null;
    oldroot = null;
  }
  
  /***************************************************************************
   * Extract formula tree (deleting the previous one), may return null
   **************************************************************************/
  public static Node.Application extract()
  {
    if (root == null) return null;
    Node.Application result = root;
    current = root = null;
    return result;
  }
  
  /***************************************************************************
   * Get number of nodes (-1, if no graph)
   **************************************************************************/
  public static int nodeNumber() 
  { 
    if (root == null) return -1;
    return nodeNumber(root);
  }
  
  /***************************************************************************
   * Get number of nodes in tree.
   * @param root the root of the tree.
   * @return the number of nodes.
   **************************************************************************/
  private static int nodeNumber(Node root)
  {
    int result = 1;
    for (Node child : root.children)
    {
      if (child instanceof Node.Application)
        result += 1;
      else
        result += nodeNumber(child);
    }
    return result;
  }
  
  /***************************************************************************
   * print tree to stdout ("no tree", if none)
   **************************************************************************/
  public static void print()
  {
    if (root == null) 
    {
      System.out.println("(no tree)");
      return;
    }
    root.print(0);
  }
  
  /***************************************************************************
   * Create a new child of the current node and make it current node.
   **************************************************************************/
  public static void enter(Context context, String label, AST phrase, Cut cut)
  {
    if (current == null) return;
    Node node = new Node(current, context, label, phrase, cut);
    current = node;
  }
  
  /***************************************************************************
   * Switch to next branch
   * @param phrase the phrase of the branch
   * @param context the context of the branch
   * @param value the value of the branch
   **************************************************************************/
  public static void next(AST phrase, Context context, boolean value)
  {
    if (current == null) return;
    while (current instanceof Node.Parent) 
    {
      current.close(value);
      current = current.parent;
    }
    current.next(phrase, context, value);
  }
  
  /***************************************************************************
   * Return to the parent (cutting branches from current node, if necessary)
   * @value the value of the current node
   **************************************************************************/
  public static void exit(boolean value)
  {
    if (current == null) return;
    current.close(value);
    do { current = current.parent; } while (current instanceof Node.Parent);
  }
  
  /***************************************************************************
   * Return to the parent (cutting branches from current node, if necessary)
   * @value the value of the current node
   **************************************************************************/
  public static void exit(Value value)
  {
    if (current == null) return;
    if (!(current instanceof Node.Application)) 
    { 
      if (value instanceof Value.Bool)
        exit(((Value.Bool)value).getValue());
      else
        exit(false);
      return;
    }
    Node.Application current0 = (Node.Application)current;
    current0.close(value);
    do { current = current.parent; } while (current instanceof Node.Parent);
  }
  
  /***************************************************************************
   * Create a new child of the current node and make it current node.
   **************************************************************************/
  public static void application(AST phrase, Context context)
  {
    if (current == null) return;
    if (!(current instanceof Node.Parent))
      current = new Node.Parent(current, context, null);
    Node node = new Node.Application(current, context, phrase.toString(), phrase, Cut.None);
    current = node;
  }
  
  /***************************************************************************
   * Create a new child of the current node and make it current node.
   **************************************************************************/
  public static void arguments(Context context, 
    ValueSymbol[] params, Value[] values, AST phrase)
  {
    if (current == null) return;
    Node node = new Node.Arguments(current, context, "(arg)", phrase, Cut.None, 
        params, values);
    current = node;
  }
  
  /***************************************************************************
   * Return to the grandparent (cutting branches from current node if necessary)
   **************************************************************************/
  public static void result(Value result)
  {
    if (current == null) return;
    while (current instanceof Node.Parent) current = current.parent;
    if (!(current instanceof Node.Arguments)) return;
    Node.Arguments argnode = (Node.Arguments)current;
    argnode.setResult(result);
    current = argnode.parent;
    if (current == root) return;
    if (!(current instanceof Node.Application)) return;
    Node.Application appnode = (Node.Application)current;
    appnode.setResult(result);
    current = appnode.parent;
  }
}
// ----------------------------------------------------------------------------
// end of file
// ----------------------------------------------------------------------------