// ---------------------------------------------------------------------------
// MapTranslator.java
// A class for the translation of various map expressions to SMT-LIB strings.
//
// Author: Franz Reichl <franz.x.reichl@gmail.com>
// $Id: MapTranslator.java,v 1.4 2020/07/02 10:15:59 schreine Exp $
// 
// Author: Franz Reichl <franz.x.reichl@gmail.com>
// Copyright (C) 2019-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ---------------------------------------------------------------------------
package riscal.smt;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map.Entry;
import java.util.function.Function;

import riscal.semantics.Types.Value;
import riscal.semantics.Types.Value.Array;
import riscal.semantics.Types.Value.Map;
import riscal.smt.ExpressionTranslator.TypeEnumerationInformation;
import riscal.smt.ExpressionTranslator.ValueContainer;
import riscal.syntax.AST;
import riscal.syntax.AST.Expression;
import riscal.syntax.AST.Expression.ApplicationExp;
import riscal.syntax.AST.Expression.EqualsExp;
import riscal.syntax.AST.Expression.IdentifierExp;
import riscal.syntax.AST.Expression.IfThenElseExp;
import riscal.syntax.AST.Expression.LetParExp;
import riscal.syntax.AST.Expression.MapBuilderExp;
import riscal.syntax.AST.Expression.MapSelectionExp;
import riscal.syntax.AST.Expression.MapUpdateExp;
import riscal.syntax.AST.Expression.NotEqualsExp;
import riscal.syntax.AST.Expression.TupleSelectionExp;
import riscal.syntax.AST.Type;
import riscal.syntax.AST.Type.BoolType;
import riscal.syntax.AST.Type.IntType;
import riscal.syntax.AST.Type.MapType;
import riscal.syntax.AST.Type.SetType;
import riscal.syntax.ASTVisitor;
import riscal.util.Seq;
import riscal.util.Seq.Next;

public class MapTranslator extends ASTVisitor.Base<String>
{
	/**
	 * The underlying ExpressionTranslator
	 */
	private ExpressionTranslator expTrans;
	
	/**
	 * Assigns to the already encountered MapTypes a MapInformation
	 */
	private HashMap<Type, MapInformation> visitedMapTypes=new HashMap<AST.Type, MapTranslator.MapInformation>();
	
	/**
	 * The basis for the generation of the name of the map access function
	 */
	private static final String defaultAccessName="mapAccess";
	/**
	 * The basis for the generation of the name of the map update function
	 */
	private static final String defaultUpdateName="mapUpdate";
	/**
	 * The basis for the generation of the name of the enumeration of a MapType.
	 */
	private static final String defaultTypeEnumerationName="mapEnum";
	
	/***************************************************************************
	 * Create a MapTranslator with the given underlying ExpressionTranslator.
	 * @param et The underlying ExpressionTranslator.
	 **************************************************************************/
	public MapTranslator(ExpressionTranslator et)
	{
		this.expTrans=et;
	}
	
	/***************************************************************************
	 * Translate the given expression to SMT-LIB.
	 * @param exp The expression that shall be translated.
	 * Has to be a map related expression.
	 * @return The result of the translation.
	 **************************************************************************/
	public String compute(Expression exp)
	{
		if(exp.getType() instanceof MapType)
		{
			storeType((MapType)exp.getType());
		}
		return exp.accept(this);
	}
	
	/***************************************************************************
	 * Translate the given expression to SMT-LIB.
	 * The expression is treated like it had the type mT.
	 * @param exp The expression that shall be translated.
	 * Has to be a map expression.
	 * @param mT Must contain the type of exp. Shall be canonical.
	 * @return The result of the translation.
	 **************************************************************************/
	public String compute(Expression exp, MapType mT)
	{
		storeType(mT);
		MapInformation info1=visitedMapTypes.get(mT);
		if(exp instanceof MapBuilderExp)
		{
			return processMapBuilder((MapBuilderExp) exp, mT.type2, info1.numberOfComponents);
		}
		String str=exp.accept(this);
		MapType old=(MapType) Utilities.getBaseType(exp.getType());
		return adaptTranslation(str, old, mT);	
	}
	
	/***************************************************************************
	 * Takes a SMT-LIB expression that has the sort that corresponds
	 * to oldType. The method transforms this SMT-LIB expression to
	 * a SMT-LIB expression of the sort that corresponds to newType.
	 * Condition:
	 * oldType must be contained in newType
	 * @param s The SMT-LIB expression that shall be transformed.
	 * @param oldType The type that corresponds to the current sort.
	 * @param newType The type that corresponds to the sort the result
	 * shall have.
	 * @return A SMT-LIB expression of the sort that corresponds to
	 * newType
	 **************************************************************************/
	String adaptTranslation(String s,MapType oldType,MapType newType)
	{
		storeType(newType);
		MapInformation info=visitedMapTypes.get(oldType);
		assert(visitedMapTypes.get(newType).numberOfComponents==info.numberOfComponents);
		long accumulatedLength=0;
		long len=info.componentLength;
		String aux="((_ extract "+Long.toString(len-1)+" "+Long.toString(accumulatedLength)+") "+s+")";
		String res=expTrans.adaptTranslation(aux, oldType.type2, newType.type2);
		if(res.equals(aux))//in this case no changes need to be done
			return s;
		accumulatedLength=len;
		for(int i=1;i<info.numberOfComponents;i++)
		{
			len=accumulatedLength+info.componentLength;
			aux="((_ extract "+Long.toString(len-1)+" "+Long.toString(accumulatedLength)+") "+s+")";
			res="(concat "+expTrans.adaptTranslation(aux, oldType.type2, newType.type2)+" "+res+")";
			accumulatedLength+=info.componentLength;
		}
		return res;
	}
	
	/***************************************************************************
	 * If the MapInformation is not available for the
	 * given type compute it.
	 * This includes the generation of function definitions for
	 * a map update and a map access function.
	 * @param mT The type of interest. Has to be canonical.
	 **************************************************************************/
	private void storeType(MapType mT)
	{
		MapInformation info=visitedMapTypes.get(mT);
		if(info!=null)
			return;
		String updateName=expTrans.getFunctionName(defaultUpdateName);
		String accessName=expTrans.getFunctionName(defaultAccessName);
		Type domainType=mT.type1;
		Type imageType=mT.type2;
		long cardinalityDomain=0;
		String enumFunName=null;
		boolean treatAsArray=false;
		if(domainType instanceof IntType)
		{
			IntType iT=(IntType)domainType;
			if(iT.ivalue1==0)
			{
				treatAsArray=true;
				cardinalityDomain=iT.ivalue2+1;
			}
		}
		if(!treatAsArray)
		{
			ExpressionTranslator.TypeEnumerationInformation enumInfo=expTrans.getTypeEnumerationFunction(domainType);
			cardinalityDomain=enumInfo.numberOfElements;
			enumFunName=enumInfo.functionName;
		}
		long indexVectorLength=IntTranslator.determineBitVectorLength(0, cardinalityDomain-1);
		long imageVectorLength=expTrans.giveVectorLength(imageType);
		info=new MapInformation(accessName, updateName, enumFunName, treatAsArray, imageVectorLength, cardinalityDomain,indexVectorLength);
		visitedMapTypes.put(mT, info);
		String mapVariable="m";
		String newValue="val";
		String indexVariable="i";
		String accDef=giveAccessDefinitionString(info,mapVariable,indexVariable);
		if(Utilities.getBaseType(imageType)instanceof BoolType)
		{
			accDef="("+expTrans.getBitVector2boolean()+" "+accDef+")";
			String newValueBool="("+expTrans.getBoolean2BitVector()+" "+newValue+")";
			String upDef=giveUpdateDefinitionString(info,mapVariable,indexVariable,newValueBool);
			expTrans.addDeclaration("(define-fun "+accessName+" (("+mapVariable+"(_ BitVec "+Long.toString(info.vectorLength)+"))"
					+ "("+indexVariable+"(_ BitVec "+Long.toString(info.indexVectorLength)+"))) "
							+ "Bool"+accDef+")");
			expTrans.addDeclaration("(define-fun "+updateName+" (("+mapVariable+"(_ BitVec "+Long.toString(info.vectorLength)+"))"
					+ "("+indexVariable+"(_ BitVec "+Long.toString(info.indexVectorLength)+"))"
					+ "("+newValue+" Bool)) (_ BitVec "+Long.toString(info.vectorLength)+")"
					+ upDef+")");		
		}
		else
		{
			String upDef=giveUpdateDefinitionString(info,mapVariable,indexVariable,newValue);
			expTrans.addDeclaration("(define-fun "+accessName+" (("+mapVariable+"(_ BitVec "+Long.toString(info.vectorLength)+"))"
					+ "("+indexVariable+"(_ BitVec "+Long.toString(info.indexVectorLength)+"))) "
							+ "(_ BitVec "+Long.toString(info.componentLength)+")"+accDef+")");
			expTrans.addDeclaration("(define-fun "+updateName+" (("+mapVariable+"(_ BitVec "+Long.toString(info.vectorLength)+"))"
					+ "("+indexVariable+"(_ BitVec "+Long.toString(info.indexVectorLength)+"))"
					+ "("+newValue+"(_ BitVec "+Long.toString(info.componentLength)+"))) (_ BitVec "+Long.toString(info.vectorLength)+")"
					+ upDef+")");		
		}		
	}	

	/***************************************************************************
	 * Generate the defining expression for the definition of the function representing the map access function.
	 * @param info The MapInformation of the type for which the access function shall be defined.
	 * @param mapVariable The name of the function argument representing the map that shall be accessed.
	 * @param indexVariable The name of the function argument that determines the
	 * position of the access.
	 * @return The defining SMT-LIB expression for the representation of the map access.
	 **************************************************************************/
	private String giveAccessDefinitionString(MapInformation info,String mapVariable,String indexVariable)
	{
		if(info.numberOfComponents==1)
			return mapVariable;
		String accDef="(ite (= "+indexVariable+" "+IntTranslator.generateSMT_LIB_Constant(0, info.indexVectorLength)+")((_ extract "
				+ Long.toString(info.componentLength-1)+" 0) "+mapVariable+")"+IntTranslator.generateSMT_LIB_Constant(0, info.componentLength)+")"; 
		long start=info.componentLength;
		for(int i=1;i<info.numberOfComponents;i++)
		{
			accDef="(ite (= "+indexVariable+" "+IntTranslator.generateSMT_LIB_Constant(i, info.indexVectorLength)+")((_ extract "
					+Long.toString(start+info.componentLength-1)+" "+start+") "+mapVariable+")"+accDef+")";
			start+=info.componentLength;
		}
		return accDef;
	}	

	/***************************************************************************
	 * Generate the defining expression for the definition of the function representing the map update function.
	 * @param info The MapInformation of the type for which the access function shall be defined.
	 * @param mapVariable The name of the function argument representing the map that shall be accessed.
	 * @param indexVariable The name of the function argument that determines the
	 * position of the access.
	 * @param newValue The name of the function argument representing the new value.
	 * @return The defining SMT-LIB expression for the representation of the map update.
	 **************************************************************************/
	private String giveUpdateDefinitionString(MapInformation info,String mapVariable,String indexVariable,String newValue)
	{
		if(info.numberOfComponents==1)
			return newValue;
		String upDef="(ite (= "+indexVariable+" "+IntTranslator.generateSMT_LIB_Constant(0, info.indexVectorLength)+")"
				+ "(concat ((_ extract "+Long.toString(info.vectorLength-1)+" "+Long.toString(info.componentLength)+")"+mapVariable+")"+newValue+")"
				+ IntTranslator.generateSMT_LIB_Constant(0, info.vectorLength)+")";
		long start=info.componentLength;
		for(int i=1;i<info.numberOfComponents-1;i++)
		{
			String aux="(concat "+newValue+"((_ extract "+Long.toString(start-1)+" "+Long.toString(0)+")"+mapVariable+"))";
			aux="(concat ((_ extract "+Long.toString(info.vectorLength-1)+" "+Long.toString(start+info.componentLength)+")"+mapVariable+") "+aux+")";
			upDef="(ite (= "+indexVariable+" "+IntTranslator.generateSMT_LIB_Constant(i, info.indexVectorLength)+")"+aux+" "+upDef+")";
			start+=info.componentLength;
		}
		upDef="(ite (= "+indexVariable+" "+IntTranslator.generateSMT_LIB_Constant(info.numberOfComponents-1, info.indexVectorLength)+")"
				+ "(concat "+newValue+"((_ extract "+Long.toString(start-1)+" "+Long.toString(0)+")"+mapVariable+"))"
				+ upDef+")";
		return upDef;
	}
	
	/***************************************************************************
	 * Compute the SMT-LIB representations of the values of the given type.
	 * Add the #b prefix to these representations depending on bare.
	 * @param mT The type whose values shall be computed. Has to be canonical.
	 * @param bare If true do not add the #b prefix to the resulting values.
	 * @return A list containing all values of the given type 
	 * represented by their corresponding SMT-LIB strings.
	 **************************************************************************/
	public List<String> computeValues(MapType mT,boolean bare)
	{
		storeType(mT);
		MapInformation info=visitedMapTypes.get(mT);
		if(info.vectorLength>Integer.MAX_VALUE)
		{
			SMT.tooLargeType("The given type is too large", info.vectorLength, mT);
		}
		int len=(int)info.vectorLength;
		int typeSize=ExpressionTranslator.getTypeSize(mT);
		int domainSize=ExpressionTranslator.getTypeSize(mT.type1);
		int imageSize=ExpressionTranslator.getTypeSize(mT.type2);
		List<String> vals=expTrans.computeValuesBare(mT.type2);
		List<String> mapVals=new ArrayList<String>(typeSize);
		//idea based on the generation of values in Values.java
		int[] counter=new int[domainSize];
		for(int i=0;i<typeSize;i++)
		{
			int carry = 1;
			for(int j=0;j<domainSize;j++)
			{
				counter[j] = counter[j] + carry;
				if(counter[j]==imageSize)
				{
					counter[j]=0;
					carry = 1;
				}
				else
					carry=0;
			}
			StringBuilder strB=new StringBuilder(len);
			if(!bare)
				strB.append("#b");
			for(int j=domainSize-1;j>=0;j--)
			{
				strB.append(vals.get(counter[j]));
			}
			mapVals.add(strB.toString());
		}
		return mapVals;
	}
	
	/***************************************************************************
	 * Translate a sequence of values with base type type to a list of SMT-LIB values.
	 * Add the #b prefix to these representations depending on bare.
	 * @param type The base type of the values. Must be canonical.
	 * @param seq A sequence of values whose basetype is type.
	 * @param bare If true do not add the #b prefix to the resulting values.
	 * @return A list containing the SMT-LIB representations of the
	 * values in seq.
	 **************************************************************************/
	List<String> processValueSequence(Type type,Seq<Value> seq,boolean bare)
	{
		type=Utilities.getBaseType(type);
		MapType mT=(MapType)type;
		storeType(mT);
		ValueContainer valC=ExpressionTranslator.ValueContainerGenerator.getValueContainer(expTrans, type);
		int len=ExpressionTranslator.getTypeSize(type);
		List<String> result=new ArrayList<String>(len);
		Function <Value, String> proc;
		if(bare)
			proc=v -> valC.processToBareVector(v);
		else
			proc=v -> valC.process(v);
		while (true)
    {
      Next<Value> next = seq.get();
      if (next == null) break;
      result.add(proc.apply(next.head));
      seq = next.tail;
    }
		return result;
	}

	/***************************************************************************
	 * Translate the given value to a bit vector constant.
	 * @param val A map value that shall be translated.
	 * @param img A ValueContainer for processing the values
	 * of the type of the image of the map.
	 * @param isArray if true the domain of the respective map type must
	 * be an IntType with a lower bound of 0.
	 * @return A SMT-LIB expression representing val.
	 **************************************************************************/
	private static String processValue(Value val, ValueContainer img,boolean isArray)
	{
		if(isArray)
			return "#b"+processValueArrayBareVector(val,img);
		else
			return "#b"+processValueMapBareVector(val,img);
	}
	
	/***************************************************************************
	 * Translate the given value to a bit sequence.
	 * @param val A map value that shall be translated.
	 * The value has to correspond to a MapType whose domain is an
	 * IntType with the lower bound 0.
	 * @param img A ValueContainer for processing the values
	 * of the type of the image of the map.
	 * @return A bit sequence representing val.
	 **************************************************************************/
	private static String processValueArrayBareVector(Value val,ValueContainer img)
	{			
		Array v=(Array)val;
		Iterator<Value> it=v.iterator();
		String str=img.processToBareVector(it.next());//There must be at least one element ->hasNext() must be true
		while(it.hasNext())
		{
			str=img.processToBareVector(it.next())+str;
		}
		return str;
	}
	
	/***************************************************************************
	 * Translate the given value to a bit sequence.
	 * @param val A map value that shall be translated.
	 * @param img A ValueContainer for processing the values
	 * of the type of the image of the map.
	 * @return A bit sequence representing val.
	 **************************************************************************/
	private static String processValueMapBareVector(Value val,ValueContainer img)
	{			
		Map v=(Map)val;
		Iterator<Entry<Value,Value>> it=v.entrySet().iterator();
		String str=img.processToBareVector(it.next().getValue());//There must be at least one element ->hasNext() must be true
		while(it.hasNext())
		{
			str=img.processToBareVector(it.next().getValue())+str;
		}
		return str;
	}
	
	
	static class MapValueContainer extends ExpressionTranslator.ValueContainer
	{
		/**
		 * A ValueContainer for processing the values of the image of the
		 * respective MapType.
		 */
		private ValueContainer imageValueContainer;
		/**
		 * If true the domain of the MapType must be an IntType with
		 * lower bound 0.
		 */
		private boolean isArray;
		MapValueContainer(ExpressionTranslator expTrans,MapTranslator mTranslator,MapType mT)
		{
			mTranslator.storeType(mT);
			imageValueContainer=ExpressionTranslator.ValueContainerGenerator.getValueContainer(expTrans, mT.type2);
			if(mTranslator.visitedMapTypes.get(mT).treatAsArray)
				isArray=true;
			else
				isArray=false;
		}
		/***************************************************************************
		 * Generate a bit vector constant representing the given value.
		 * @param val The value that shall be represented by a bit vector.
		 * @return The SMT-LIB representation of val
		 **************************************************************************/
		String process(Value val) 
		{
			return processValue(val, imageValueContainer, isArray);
		}
		/***************************************************************************
		 * Generate a bit sequence representing the given value.
		 * @param val The value that shall be represented by a bit vector.
		 * @return A bit sequence representing val.
		 **************************************************************************/
		String processToBareVector(Value val)
		{
			if(isArray)
			{
				return processValueArrayBareVector(val, imageValueContainer);
			}
			else
			{
				return processValueMapBareVector(val, imageValueContainer);
			}
		}
	}
	
	/***************************************************************************
	 * Generate an enumeration function for the given type and return the
	 * associated information.
	 * @param mT The type for which an enumeration function shall be generated.
	 * @return The information associated to the enumeration function.
	 **************************************************************************/
	public TypeEnumerationInformation generateTypeEnumerationFunction(MapType mT)
	{
		String funName=expTrans.getFunctionName(defaultTypeEnumerationName);
		storeType(mT);
		MapInformation info=visitedMapTypes.get(mT);
		long typeSize=ExpressionTranslator.getTypeSize(mT);
		int resultVectorLength=IntTranslator.determineBitVectorLength(0, typeSize-1);
		String variable="x";
		//A Map with e.g. a unit image would make problems if not treated separately
		if(typeSize==1)
		{
			String def="(define-fun "+funName+" (("+variable+" (_ BitVec "+Long.toString(info.vectorLength)+")))"
					+ "(_ BitVec "+Long.toString(resultVectorLength)+") "+
					IntTranslator.generateSMT_LIB_Constant(0, resultVectorLength)+")";
			expTrans.addDeclaration(def);
			return new TypeEnumerationInformation(funName, resultVectorLength, typeSize);
		}
		TypeEnumerationInformation enumInfo=expTrans.getTypeEnumerationFunction(mT.type2);
		int diff=resultVectorLength-enumInfo.enumerationFunctionResultVectorLength;
		if(diff<0)
			SMT.illegalUsage(null, "Enumeration of Maps - erroneous vector length of components");	
		long accumulatedSize=1;
		if(info.numberOfComponents>Integer.MAX_VALUE)
			SMT.illegalUsage(null, "Map with too many components");
		List<String> slst=new ArrayList<String>((int) info.numberOfComponents);
		for(int i=0;i<info.numberOfComponents;i++)
		{
			long lower=i*info.componentLength;
			long upper=(i+1)*info.componentLength-1;
			String subVector="((_ extract "+Long.toString(upper)+" "+Long.toString(lower)+")"+variable+")";
			subVector="((_ zero_extend "+Long.toString(diff)+") ("+enumInfo.functionName+" "+subVector+"))";
			if(accumulatedSize>1)
				subVector="(bvmul "+IntTranslator.generateSMT_LIB_Constant(accumulatedSize, resultVectorLength)+" "+subVector+")";
			slst.add(subVector);
			accumulatedSize*=enumInfo.numberOfElements;
		}
		String args=String.join(" ", slst);
		if(info.numberOfComponents>1)
			args="(bvadd "+args+")";
		String def="(define-fun "+funName+" (("+variable+" (_ BitVec "+Long.toString(info.vectorLength)+")))"
				+ "(_ BitVec "+Long.toString(resultVectorLength)+") "+args+")";
		expTrans.addDeclaration(def);
		return new TypeEnumerationInformation(funName, resultVectorLength, typeSize);
	}
	
	/***************************************************************************
	 * Generates an enumeration function for the values given by lst.
	 * The result of the enumeration is given by the position 
	 * of the respective value in the list.
	 * The enumeration function is defined via a table.
	 * @param lst The values used to generate the enumeration function.
	 * @param mT The elements of lst must correspond to the type mT. Must be canonical.
	 * @return The information associated to the enumeration function.
	 **************************************************************************/
	public TypeEnumerationInformation generateTypeEnumerationFunctionUseTable(List<Value> lst,MapType mT)
	{
		String funName=expTrans.getFunctionName(defaultTypeEnumerationName);
		storeType(mT);
		MapInformation info=visitedMapTypes.get(mT);
		long len=lst.size();
		long typeVectorLength=info.vectorLength;
		int argumentVectorLength=IntTranslator.determineBitVectorLength(0, len-1);
		expTrans.addDeclaration("(declare-fun "+funName+" ((_ BitVec "+Long.toString(typeVectorLength)+")) "
				+ "(_ BitVec "+Long.toString(argumentVectorLength)+"))");
		ValueContainer valC=ExpressionTranslator.ValueContainerGenerator.getValueContainer(expTrans, mT);
		long i=0;		
		for(Value val:lst)
		{
			expTrans.addDeclaration("(assert (= ("+funName+" "+valC.process(val)+") "+
					IntTranslator.generateSMT_LIB_Constant(i, argumentVectorLength)+"))");
			i++;
		}		
		return new TypeEnumerationInformation(funName, argumentVectorLength, len);
	}		
	
	public TypeEnumerationInformation generateTypeEnumerationFunctionUseTableWithFunctionDefinition(List<Value> lst,MapType mT)
	{
		String funName=expTrans.getFunctionName(defaultTypeEnumerationName);
		storeType(mT);
		MapInformation info=visitedMapTypes.get(mT);
		long len=lst.size();
		long typeVectorLength=info.vectorLength;
		int argumentVectorLength=IntTranslator.determineBitVectorLength(0, len-1);
		ValueContainer valC=ExpressionTranslator.ValueContainerGenerator.getValueContainer(expTrans, mT);
		Iterator<Value> it=lst.iterator();
		StringBuilder strB=new StringBuilder();
		String variable="x";
		for(int i=0;i<len-1;i++)
		{
			Value val=it.next();
			strB.append("(ite (= "+variable+" "+valC.process(val)+")"+
					IntTranslator.generateSMT_LIB_Constant(i, argumentVectorLength)+" ");
		}
		strB.append(IntTranslator.generateSMT_LIB_Constant(len-1, argumentVectorLength));
		strB.append(")".repeat((int) (len-1)));
		String def="(define-fun "+funName+" (("+variable+" (_ BitVec "+Long.toString(typeVectorLength)+"))) "
				+ "(_ BitVec "+Long.toString(argumentVectorLength)+") "+strB.toString()+")";
		expTrans.addDeclaration(def);
		return new TypeEnumerationInformation(funName, argumentVectorLength, len);
	}
		
	/***************************************************************************
	 * Compute the bit vector length that is necessary to represent the values
	 * of the given type.
	 * @param mT The type of interest. Must be canonical.
	 * @return The necessary length to represent the values of mT.
	 **************************************************************************/
	public long giveVectorLength(MapType mT)
	{
		storeType(mT);
		MapInformation info=visitedMapTypes.get(mT);
		return info.vectorLength;
	}
	
	/***************************************************************************
	 * If not yet generated, generate the function representing the map access
	 * for the given MapType. Apply this function to the SMT-LIB expression
	 * map that represents a map expression and to the bit vector representation
	 * of position.
	 * @param map A bit vector expression representing a map expression.
	 * @param mT The MapType of the expression that is represented by map.
	 * @param position Determines the position where the map shall be accessed.
	 * @return The application of the representation of the map access.
	 **************************************************************************/
	private String generateMapAccess(String map,MapType mT,Expression position)
	{		
		storeType(mT);
		Type domainType=mT.type1;
		MapInformation info=visitedMapTypes.get(mT);
		String index;
		if(info.treatAsArray)
		{
			IntType iT=new IntType(0, info.numberOfComponents-1);
			iT.setCanonical();//maybe unnecessary
			index=expTrans.compute(position, iT);
		}
		else
		{
			index=expTrans.compute(position, domainType);
			index="("+info.enumerationFunctionName+" "+index+")";
		}
		String acc="("+info.accessName+" "+map+" "+index+")";
		return acc;		
	}
	
	/***************************************************************************
	 * If not yet generated, generate the function representing the map update
	 * for the given MapType.
	 * Apply this function to the bit vector representation of map,
	 * to the bit vector representation
	 * of position and to the bit vector representation of newValue.
	 * @param map The map expression that shall be updated
	 * @param position Determines the position where the map shall be accessed.
	 * @param newValue Determines the value that shall be assigned to the map
	 * at the respective position.
	 * @return The application of the representation of the map update.
	 **************************************************************************/
	private String generateMapUpdate(Expression map,Expression position,Expression newValue)
	{
		MapType mT=(MapType)map.getType();
		Type domainType=mT.type1;
		storeType(mT);
		MapInformation info=visitedMapTypes.get(mT);
		String index;
		if(info.treatAsArray)
		{
			IntType iT=new IntType(0, info.numberOfComponents-1);
			iT.setCanonical();//maybe unnecessary
			index=expTrans.compute(position, iT);
		}
		else
		{
			index=expTrans.compute(position, domainType);
			index="("+info.enumerationFunctionName+" "+index+")";
		}
		String mapStr=compute(map);
		String val=expTrans.compute(newValue,mT.type2);
		String up="("+info.updateName+" "+mapStr+" "+index+" "+val+")";
		return up;
	}
	
	/***************************************************************************
	 * Generates a boolean valued SMT-LIB expression that is true
	 * iff the SMT-LIB expression represents a bit vector value
	 * that represents a value of the given MapType.
	 * @param mT The type of interest.
	 * @param mapStr A bit vector expression.
	 * @return A boolean valued SMT-LIB expression.
	 **************************************************************************/
	public String generateTypeConditions(MapType mT,String mapStr)
	{
		Type imageType=mT.type2;
		if(imageType instanceof SetType)
			return null;
		List<String> lst=new ArrayList<String>();
		storeType(mT);
		MapInformation info=visitedMapTypes.get(mT);
		for(int i=0;i<info.numberOfComponents;i++)
		{
			String index=IntTranslator.generateSMT_LIB_Constant(i, info.indexVectorLength);
			String acc="("+info.accessName+" "+mapStr+" "+index+")";					
			String s=expTrans.generateTypeConditions(imageType,acc);
			if(s!=null)
				lst.add(s);
		}
		if(lst.size()==0)
			return null;
		else if(lst.size()==1)//Boolector does not support and with one argument
		{
			return String.join(" ", lst);
		}
		else
		{
			return "(and "+String.join(" ", lst)+")";
		}
	}
	
	public String visit(IdentifierExp exp)
	{
		return expTrans.compute(exp);
	}
	
	public String visit(ApplicationExp exp)
	{
		return expTrans.compute(exp);
	}
	
	public String visit(EqualsExp exp)
	{
		MapType mT1=(MapType)exp.exp1.getType();
		MapType mT2=(MapType)exp.exp2.getType();
		storeType(mT1);
		storeType(mT2);
		MapInformation info1=visitedMapTypes.get(mT1);
		MapInformation info2=visitedMapTypes.get(mT2);
		if(info1.numberOfComponents!=info2.numberOfComponents)
			return "false";
		
		Utilities.Pair<String, String> pair=expTrans.computeUseSameType(exp.exp1, exp.exp2);
		return "(= "+pair.first+" "+pair.second+")";
	}
	
	public String visit(NotEqualsExp exp)
	{
		MapType mT1=(MapType)exp.exp1.getType();
		MapType mT2=(MapType)exp.exp2.getType();
		storeType(mT1);
		storeType(mT2);
		MapInformation info1=visitedMapTypes.get(mT1);
		MapInformation info2=visitedMapTypes.get(mT2);
		if(info1.numberOfComponents!=info2.numberOfComponents)
			return "true";
		
		Utilities.Pair<String, String> pair=expTrans.computeUseSameType(exp.exp1, exp.exp2);
		return "(distinct "+pair.first+" "+pair.second+")";
	}
	
	public String visit(MapBuilderExp exp)
	{
		MapType mT=(MapType)exp.getType();
		storeType(mT);
		MapInformation info=visitedMapTypes.get(mT);
		return processMapBuilder(exp, mT.type2, info.numberOfComponents);
	}
	
	/***************************************************************************
	 * Create the SMT-LIB representation of a MapBuilder expression
	 * @param exp The MapBuilder of interest.
	 * @param componentType The type of the image of respective MapType.
	 * @param numberOfComponents The size of the domain of the MapType.
	 * @return numberOfComponents repetitions of the SMT-LIB representation
	 * of exp.exp. 
	 **************************************************************************/
	private String processMapBuilder(MapBuilderExp exp,Type componentType,long numberOfComponents)
	{
		String elem=expTrans.compute(exp.exp,componentType);
		if(Utilities.getBaseType(componentType)instanceof BoolType)
		{
			elem="("+expTrans.getBoolean2BitVector()+" "+elem+")";
		}
		return "((_ repeat "+Long.toString(numberOfComponents)+")"+elem+")";
	}
	
	public String visit(MapSelectionExp exp)
	{
		String map=compute(exp.exp1);
		MapType mT=(MapType)exp.exp1.getType();
		return generateMapAccess(map, mT, exp.exp2);
	}
	
	public String visit(MapUpdateExp exp)
	{
		return generateMapUpdate(exp.exp1, exp.exp2, exp.exp3);
	}
	
	public String visit(LetParExp exp)
	{
		return expTrans.compute(exp);
	}
	
	public String visit(TupleSelectionExp exp)
	{
		return expTrans.compute(exp);
	}
	
//	public String visit(IfThenElseExp exp)
//	{
//		String cond=expTrans.compute(exp.exp1);
//		String thenExp=exp.exp2.accept(this);
//		String elseExp=exp.exp3.accept(this);
//		return "(ite "+cond+" "+thenExp+" "+elseExp+")";
//	}	
	
	public String visit(IfThenElseExp exp)
	{
		return expTrans.compute(exp);
	}
	
	public String visit(MapType mT)
	{
		assert(mT.isCanonical());
		storeType(mT);
		MapInformation info=visitedMapTypes.get(mT);		
		return "(_ BitVec "+Long.toString(info.vectorLength)+")";
	}
	
	/**
	 * Stores various information related to a MapType.
	 */
	private class MapInformation
	{
		public String accessName;
		public String updateName;
		public String enumerationFunctionName;
		boolean treatAsArray;
		public long componentLength;
		public long numberOfComponents;
		public long indexVectorLength;
		public long vectorLength;
		public MapInformation(String ac,String up,String enumName,boolean array,long comp,long numComps,long indexLength)
		{
			this.accessName=ac;
			this.updateName=up;
			this.enumerationFunctionName=enumName;
			this.treatAsArray=array;
			this.componentLength=comp;
			this.numberOfComponents=numComps;
			this.indexVectorLength=indexLength;
			this.vectorLength=comp*numComps;
		}
	}
}
//---------------------------------------------------------------------------
// end of file
//---------------------------------------------------------------------------
