// ---------------------------------------------------------------------------
// TypeBounds.java
// Replace expressions in type bounds by newly introduced named constants.
// $Id: TypeBounds.java,v 1.3 2022/06/29 16:07:13 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.jku.at>
// Copyright (C) 2022-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, https://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package riscal.syntax;

import java.util.*;

import riscal.syntax.AST.*;
import riscal.syntax.AST.Type.*;
import riscal.syntax.AST.Expression.*;
import riscal.syntax.AST.Declaration.*;

public class TypeBounds extends ASTCloner
{
  public static Specification process(Specification spec)
  {
    TypeBounds processor = new TypeBounds();
    return processor.visit(spec);
  }
 
  // handling the generation of names for type bounds
  private List<Declaration> result = new ArrayList<Declaration>();
  private int nameCounter = 0;
  private static final String nameBase = "_tbound_";
  
  // remembering the previously generated declarations
  private Map<String,Identifier> ids = new HashMap<String,Identifier>();
  
  public Specification visit(Specification spec)
  {
    for (Declaration decl : spec.declarations)
    {
      Declaration decl0 = (Declaration)decl.accept(this);
      result.add(decl0);
    }
    return new Specification(result);
  }
  
  public IdentifierExp identifierExp(Expression exp)
  {
    if (exp instanceof IdentifierExp) return (IdentifierExp)exp;
    SourcePosition pos = exp.getPosition();
    String key = exp.toString();
    Identifier id = ids.get(key);
    if (id == null) 
    {
      String name = nameBase + nameCounter;
      nameCounter++;
      id = new Identifier(name);
      id.setPosition(pos);
      Type type = new IntType(exp,exp);
      ValueDefinition decl = new ValueDefinition(id, type, exp);
      result.add(decl);
    }
    IdentifierExp exp0 = new IdentifierExp(id);
    exp0.setPosition(pos);
    return exp0;
  }
  
  public Type visit(NatType type)
  {
    IdentifierExp exp = identifierExp(type.exp);
    return setPosition(type, new NatType(exp));
  }
  
  public Type visit(IntType type)
  {
    IdentifierExp exp1 = identifierExp(type.exp1);
    IdentifierExp exp2 = identifierExp(type.exp2);
    return setPosition(type, new IntType(exp1, exp2));
  }
  
  public Type visit(ArrayType type)
  {
    IdentifierExp exp = identifierExp(type.exp);
    return setPosition(type, new ArrayType(exp, type.type));
  }
}
// ---------------------------------------------------------------------------
// end of file
// ---------------------------------------------------------------------------