// ---------------------------------------------------------------------------
// Choices.java
// Replace choices from expressions (as far as possible)
// $Id: Choices.java,v 1.28 2021/06/01 12:47:01 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.jku.at>
// Copyright (C) 2018-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package riscal.tasks;

import java.util.*;

import riscal.syntax.*;
import riscal.syntax.AST.*;
import riscal.syntax.AST.Expression.*;
import riscal.types.Environment.Symbol.*;
import riscal.smt.*;

public class Choices extends ASTCloner
{
  // this is the new version of choice elimination intended for SMT evaluation
  // (no consideration of evaluation efficiency); by eliminating choices
  // we avoid the generation of uninterpreted functions and their axiomatization
  // which affects the efficiency of most SMT solvers very negatively
  //
  // NOTE: this elimination is only sound under the assumption that the
  // specified choice can be indeed made: we generate for a choice
  // (choose x:T with F(x,y)) the constraint F(x,y) rather than
  // (exists x:T with F(x,y)) => F(x,y); this assumption is in particular true 
  // if the choice arises from the contract of a (previously verified) function;
  // it is the responsibility of the user to apply this option correctly.
  //
  // the core idea is to pull out of a expressions chosen variables x1,x2,x3,...
  // as a sequence of quantified variables with constraints
  //
  //   x1:T1 with F(a,x1), x2:T2 with F(a,x1,x2), x3:T3 with F(a,x1,x2,x3), ...
  //
  // replacing the choice expressions by the variable tuple <<x1,x2,x3,...>>; if the
  // name of a chosen variable xi is already visible in the current context,
  // it is renamed to a fresh variable xi, the corresponding constraint is 
  // instantiated as "let xi=yi in F(...,xi)";
  //
  // the quantified variables that have been pulled out yield a universally quantified
  // formula on the level of the next enclosing formula (in case of a
  // non-Boolean choice expression, this is the enclosing atomic formula)
  // provided that this formula appears in a positive universal context; alternatively
  // it may yield an existential formula, if this formula appears in a negative
  // existential context; in all other quantification contexts, the elimination 
  // is not performed.
  //
  // to avoid the blocking of this process by variable binders, we also
  // treat a binding expression "let x=T in E" as a quantified expression
  // that introduces quantified variable "x:D with x=y" which is pulled to
  // the next level of quantification level (only if a choice is made in E)
  //
  //   q(x, choose y with p1(x,y), choose y with p2(x,y))
  //   -> forall y with p1(x,y), y0 with (let y=y0 in p2(x,y)). q(x,x,x0)
  //
  //   ~exists x with p(x). q(x,choose y with p1(x,y), choose y with p2(x,y))
  //   -> ~exists x with p(x), y with p1(a,y), y0 with let y = y0 in p2(x,y).
  //        q(x,y,y0)
  //

  // quick search for syntactic occurrences of possible choose expressions 
  // or functions to be inlined to avoid overhead of cloning the formula 
  // if certainly not required
  public boolean chooseSearch(Expression exp)
  {
    ChooseSearch search = new ChooseSearch();
    exp.accept(search);
    return search.result;
  }
  private class ChooseSearch extends ASTVisitor.Base<Void>
  {
    private boolean result = false;
    public Void visit(ChooseExp exp) { result = true; return null; }
    public Void visit(ChooseInExp exp) { result = true; return null; }
    public Void visit(ApplicationExp exp)
    {
      super.visit(exp);
      if (result) return null;
      if (SMT.getInlineDef()) result = true;
      return null;
    }
  }
  
  private static enum Polarity { ptrue, pfalse, punknown };
  public static Polarity pnot(Polarity p) 
  { 
    switch (p)
    {
    case ptrue: return Polarity.pfalse;
    case pfalse: return Polarity.ptrue;
    default: return Polarity.punknown;
    }
  }
  
  /***************************************************************************
   * Remove choices in body of a theorem as far as possible.
   * @param exp the body of the theorem.
   * @param callers the call map to identify recursive functions
   * @return the theorem body with a lot of choices removed.
   **************************************************************************/
  public static Expression optimizeTheorem(Expression exp,
    Recursion.CallMap callers)
  {
    // System.out.println(exp);
    Set<String> visible = FreeVariables.compute(exp);
    Expression exp0 = optimizeFormula(exp, true, visible, callers);
    // if (exp0 != exp) System.out.println(exp0);
    return exp0;
  }
   
  /***************************************************************************
   * Optimize expression (formula or term) with given polarity.
   * @param exp the expression (formula or term).
   * @param polarity the polarity.
   * @return the optimized expression.
   **************************************************************************/
  private Expression optimizeExpression(Expression exp, Polarity polarity)
  {
    if (isFormula(exp)) 
      return optimizeFormula(exp, polarity);
    else
      return optimizeTerm(exp);
  }
    
  /***************************************************************************
   * Optimize term.
   * @param exp the term.
   * @return the optimized term.
   **************************************************************************/
  private Expression optimizeTerm(Expression exp)
  {
    if (!this.chooseSearch(exp)) return exp;
    return optimize(exp);
  }
  
  /****************************************************************************
   * Optimize formula with given polarity.
   * @param exp the formula.
   * @param polarity the polarity.
   * @return the optimized formula.
   ***************************************************************************/
  private Expression optimizeFormula(Expression exp, Polarity polarity)
  {
    if (!chooseSearch(exp)) return exp;
    switch (polarity)
    {
    case ptrue: return optimizeFormula(exp, true, visible, callers);
    case pfalse: return optimizeFormula(exp, false, visible, callers);
    default: return optimize(exp);
    }
  }
  
  /***************************************************************************
   * Remove choices in a formula as far as possible.
   * @param exp the formula
   * @param polarity the polarity of the formula 
   * @param visible the set of variables visible in the formula
   * @param callers the function map to identify recursive function calls
   * @return the the formula with a lot of choices removed.
   **************************************************************************/
  private static Expression optimizeFormula(Expression exp, 
    boolean polarity, Set<String> visible, Recursion.CallMap callers)
  {
    Polarity p = polarity ? Polarity.ptrue : Polarity.pfalse;
    Choices choices = new Choices(p, visible, callers);
    if (!choices.chooseSearch(exp)) return exp;
    Expression exp0 = choices.optimize(exp); 
    return letFormula(polarity, choices.vars, exp0);
  }
  
  /***************************************************************************
   * Core of the optimization.
   * @param exp the expression.
   * @return the optimized expression.
   **************************************************************************/
  private Expression optimize(Expression exp)
  {
    return (Expression)exp.accept(this);
  }
  
  // the current polarity
  private Polarity polarity;
  
  // the set of currently visible variable names;
  private Set<String> visible;
  
  // the list of extracted variables (quantified variable cores with constraints)
  private List<QuantifiedVariableCore> vars;
  
  // the function map to identify recursive function calls
  private Recursion.CallMap callers;
  
  /***************************************************************************
   * Create object for removing choices from formula.
   * @param polarity the polarity of the formula
   * @param visible the names of the currently visible variables.
   * @param callers the function map to identify recursive function calls
   **************************************************************************/
  private Choices(Polarity polarity, Set<String> visible,
    Recursion.CallMap callers)
  {
    // also clone expression types
    super(true);
    this.polarity = polarity;
    this.visible = visible;
    this.vars = new ArrayList<QuantifiedVariableCore>();
    this.callers = callers;
  }
  
  // we have to consider all quantified expressions
  // - choices: ChooseExp, ChooseInExp, ChooseInElseExp
  // - formulas: ForallExp, ExistsExp
  // - terms: MinExp, MaxExp, NumberExp, SumExp, ProductExp, SetBuilderExp
  // - binders: LetExp, LetParExp
  // patterns (MatchExp) is currently not supported by SMT solver, thus ignored
 
  /***************************************************************************
   * Reset list of quantified variables.
   * @return the old variables.
   *************************************************************************/
  private List<QuantifiedVariableCore> resetVariables()
  {
    List<QuantifiedVariableCore> old = vars;
    vars = new ArrayList<QuantifiedVariableCore>();
    return old;
  }
  
  /***************************************************************************
   * Get (cloned) set of visible variables.
   * @return the old variables.
   *************************************************************************/
  private Set<String> getVisible()
  {
    Set<String> old = visible;
    visible = new HashSet<String>();
    return old;
  }
  
  /***************************************************************************
   * Get quantified variable from generated variable cores;
   * resets the generated variables to the old list and the
   * set of visible variables to the old set.
   * @param ovars the old list of quantified variables.
   * @param oset the old set of visible variables.
   * @return the new variable
   *************************************************************************/
  private List<QuantifiedVariableCore> getVariable(List<QuantifiedVariableCore> ovars,
    Set<String> oset)
  {
    List<QuantifiedVariableCore> result = vars;
    vars = ovars;
    visible = oset;
    return result;
  }
  
  /***************************************************************************
   * Determine whether an expression is a formula.
   * @param exp the expression.
   * @return true if the expression denotes a formula.
   **************************************************************************/
  private static boolean isFormula(Expression exp)
  {
    Type type = exp.getType();
    while (type instanceof Type.SubType)
    {
      Type.SubType type0 = (Type.SubType)type;
      type = type0.type;
    }
    return type instanceof Type.BoolType;
  }
 
  /**************************************************************************
   * Visit an expression.
   * @param exp the expression
   * @return a version of exp with choices minimized
   **************************************************************************/
  
  public Identifier visit(Identifier id)
  {
    // do not clone identifier, preserve its symbol
    return id;
  }
  
  public Expression visit(ApplicationExp exp)
  {
    List<Expression> exps0 = new ArrayList<Expression>();
    for (Expression e : exp.exps) 
    { 
      Expression e0 = setType(e, (Expression)e.accept(this));
      exps0.add(e0);
    }
    Expression exp0 = SMT.getInlineDef() ?
        transform(exp.ident, exps0) : new ApplicationExp(exp.ident, exps0);
    return setType(exp, exp0);
  }
  
  /***************************************************************************
   * Transform function application.
   * @param fident the identifier of the function.
   * @param args the arguments of the function.
   * @return the transformed expression.
   **************************************************************************/
  private Expression transform(Identifier fident, List<Expression> args)
  {
    FunctionSymbol fun = (FunctionSymbol)fident.getSymbol();
    Declaration def = fun.getDefinition();
    OpInfo info = OpInfo.getInfo(def);
    List<Identifier> params = new ArrayList<Identifier>();
    for (Parameter p : info.param) params.add(p.ident);
    
    // replace application of operation by its contract if
    // - the operation is a procedure OR
    // - OLD: the current function is called recursively by this operation OR
    // - NEW: the operation is (directly or indirectly recursive)
    // - the function is annotated as "modular"
    boolean doit = 
       def instanceof Declaration.ProcedureDefinition || 
       // (current != null && callers.calls(fun, current));
       callers.calls(fun, fun);
    if (!doit)
    {
      for (FunctionSpec s : info.spec)
      {
        doit = s instanceof FunctionSpec.ContractSpec;
        if (doit) break;
      }
    }
    if (doit) 
    {
      // replace application by choice of value satisfying the operation's postcondition
      Expression postFormula = info.postFormula;
      if (postFormula == null) 
      {
        postFormula = new Expression.TrueExp();
        postFormula.setType(new Type.BoolType());
      }
      List<QuantifiedVariableCore> qv0 = new ArrayList<QuantifiedVariableCore>();
      ValueSymbol result = fun.getResult();
      qv0.add(new QuantifiedVariableCore.IdentifierTypeQuantifiedVar(result.ident, fun.type, postFormula));
      QuantifiedVariable qvar = new QuantifiedVariable(qv0, null);
      Expression choice = new Expression.ChooseExp(qvar);
      choice.setType(fun.type);
      choice = CloneIds.bind(params, args, choice);
      return optimize(choice);
    }
    
    // replace application by (also optimized) and instantiated function body
    Expression exp = CloneIds.bind(params, args, info.bodyexp); 
    return optimize(exp);
  }
    
  public Expression visit(NegationExp exp)
  {
    if (!chooseSearch(exp)) return exp;
    return setType(exp, 
        // revert polarity of base expression     
        new NegationExp(optimizeFormula(exp.exp, pnot(polarity))));
  }
  
  public Expression visit(AndExp exp)
  {
    if (!chooseSearch(exp)) return exp;
    return setType(exp, new AndExp(
        optimizeFormula(exp.exp1, polarity),
        optimizeFormula(exp.exp2, polarity)));
  }
  
  public Expression visit(OrExp exp)
  {
    if (!chooseSearch(exp)) return exp;
    return setType(exp, new OrExp(
        optimizeFormula(exp.exp1, polarity),
        optimizeFormula(exp.exp2, polarity)));
  }
  
  public Expression visit(ImpliesExp exp)
  {
    if (!chooseSearch(exp)) return exp;
    return setType(exp, new ImpliesExp(
        // revert polarity of assumption
        optimizeFormula(exp.exp1, pnot(polarity)),
        optimizeFormula(exp.exp2, polarity)));
  }
  
  public Expression visit(EquivExp exp)
  {
    if (!chooseSearch(exp)) return exp;
    return setType(exp, new EquivExp(
        // no definite polarity for both parts
        optimizeFormula(exp.exp1, Polarity.punknown),
        optimizeFormula(exp.exp2, Polarity.punknown)));
  }
  
  public Expression visit(IfThenElseExp exp)
  {
    if (!chooseSearch(exp)) return exp;
    return setType(exp, new IfThenElseExp(
        // no definite polarity for condition
        optimizeFormula(exp.exp1, Polarity.punknown), 
        optimizeExpression(exp.exp2, polarity),
        optimizeExpression(exp.exp3, polarity)));
  }
  
  public Expression visit(ChooseExp exp)
  {
    return quantifiedVariables(exp.qvar, null);
  }
  
  public Expression visit(ChooseInExp exp)
  {
    return quantifiedVariables(exp.qvar, exp.exp);
  }
  
  public Expression visit(ChooseInElseExp exp)
  {
    // transform expression "choose x with F in E1 else E2" into expression
    // "if exists x with F. true then choose x with F in E1 else E2"
    Map<ValueSymbol,ValueSymbol> map = new HashMap<ValueSymbol,ValueSymbol>();
    for (QuantifiedVariableCore q : exp.qvar.qvcore)
    {
      Identifier ident = q.ident;
      Identifier ident0 = new Identifier(ident.string);
      ValueSymbol symbol = (ValueSymbol)ident.getSymbol();
      ValueSymbol symbol0 = new ValueSymbol(ident0, symbol.type, symbol.kind);
      ident0.setSymbol(symbol0);
      map.put(symbol, symbol0);
    }
    Expression body = new TrueExp();
    body.setType(new Type.BoolType());
    Expression exists = new ExistsExp(exp.qvar, body);
    exists.setType(body.getType());
    Expression exists0 = CloneIds.clone(exists, map);
    Expression then = new ChooseInExp(exp.qvar, exp.exp1);
    then.setType(exp.exp1.getType());
    Expression result = new IfThenElseExp(exists0, then, exp.exp2);
    result.setType(exp.exp1.getType());
    return optimize(result);
  }
  
  public Expression visit(ForallExp exp)
  {
    if (!chooseSearch(exp)) return exp;
    // positive polarity required for universal quantifiers
    if (polarity != Polarity.ptrue) return exp;
    return quantifiedFormula(true, exp.qvar, exp.exp);
  }
  
  public Expression visit(ExistsExp exp)
  {
    if (!chooseSearch(exp)) return exp;
    // negative polarity required for existential quantifiers
    if (polarity != Polarity.pfalse) return exp;
    return quantifiedFormula(false, exp.qvar, exp.exp);
  }  
  
  public Expression visit(MinExp exp)
  {
    // do not optimize within other quantified expressions
    return exp;
  } 
  
  public Expression visit(MaxExp exp)
  {
    // do not optimize within other quantified expressions
    return exp;
  } 
  
  public Expression visit(SumExp exp)
  {
    // do not optimize within other quantified expressions
    return exp;
  } 
  
  public Expression visit(ProductExp exp)
  {
    // do not optimize within other quantified expressions
    return exp;
  } 
  
  public Expression visit(NumberExp exp)
  {
    // do not optimize within other quantified expressions
    return exp;
  } 
  
  public Expression visit(SetBuilderExp exp)
  {
    // do not optimize within other quantified expressions
    return exp;
  } 
  
  public Expression visit(LetExp exp)
  {
    if (!chooseSearch(exp)) return exp;
    return setType(exp, letExpression(Arrays.asList(exp.binder), exp.exp, false));
  }
  
  public Expression visit(LetParExp exp)
  {
    if (!chooseSearch(exp)) return exp;
    return setType(exp, letExpression(Arrays.asList(exp.binder), exp.exp, true));
  }
    
  // --------------------------------------------------------------------------
  //
  // Auxiliaries
  //
  // --------------------------------------------------------------------------

  /***************************************************************************
   * Process let expression by translation to quantified variables with equalities
   * @param binder the binders of the let expression
   * @param exp the body of the let expression
   * @param parallel true if the binders bind "in parallel"
   * @return the result of the processing.
   **************************************************************************/
  private Expression letExpression(List<Binder> binder, Expression exp, 
    boolean parallel)
  {
    List<QuantifiedVariableCore> vars0 = new ArrayList<QuantifiedVariableCore>();
    for (Binder b : binder)
    {
      ValueSymbol symbol = (ValueSymbol)b.ident.getSymbol();
      Identifier ident = symbol.ident;
      Identifier ident0 = parallel && visible.contains(ident.string) ? 
          freshIdent(ident) : ident;
      Expression idexp0 = identifierExp(ident0);
      Expression bexp = optimizeExpression(b.exp, polarity); 
      Expression eq0 = new EqualsExp(idexp0, bexp);
      eq0.setType(new Type.BoolType());
      QuantifiedVariableCore var = 
          new QuantifiedVariableCore.IdentifierTypeQuantifiedVar(ident0, 
              symbol.type, eq0);
      vars.add(var);
      visible.add(ident0.string);
      if (ident != ident0) 
      {
        Expression idexp1 = identifierExp(ident);
        Expression idexp2 = identifierExp(ident0);
        Expression eq1 = new EqualsExp(idexp1, idexp2);
        eq1.setType(new Type.BoolType());
        QuantifiedVariableCore var0 = 
            new QuantifiedVariableCore.IdentifierTypeQuantifiedVar(ident, 
                symbol.type, eq1);
        vars0.add(var0);
      }
    }
    if (parallel)
    {
      for (QuantifiedVariableCore var : vars0)
      {
        vars.add(var);
        visible.add(var.ident.string);
      }
    }
    return optimizeExpression(exp, polarity);
  }
  
  /***************************************************************************
   * Construct typed identifier expression.
   * @param ident the identifier.
   * @return the expression.
   **************************************************************************/
  private static IdentifierExp identifierExp(Identifier ident)
  {
    ValueSymbol symbol = (ValueSymbol)ident.getSymbol();
    IdentifierExp result = new IdentifierExp(ident);
    result.setType(symbol.type);
    return result;
  }
  
  /****************************************************************************
   * Process quantified formula.
   * @param forall true if universal quantification, false if existential.
   * @param qvar the quantified variable.
   * @param exp the formula body.
   * @return the processed expression.
   ***************************************************************************/
  private Expression quantifiedFormula(boolean forall, 
    QuantifiedVariable qvar, Expression exp)
  {
    List<QuantifiedVariableCore> old = resetVariables();
    Set<String> oset = getVisible();
    Expression exp0 = quantifiedVariables(qvar, exp);
    List<QuantifiedVariableCore> vars = getVariable(old, oset); 
    return letFormula(forall, vars, exp0);
  }
  
  /****************************************************************************
   * Transform variables in quantified formula to let-bound variables,
   * whenever possible.
   * @param forall true if formula is universal (false, if existential)
   * @param vars the quantified variables.
   * @param exp the formula body.
   * @return the transformed formula.
   ***************************************************************************/
  private static Expression letFormula(boolean forall,
    List<QuantifiedVariableCore> vars, Expression exp)
  {
    // generate result formula (transforming quantified variables to
    // let-bound variables, whenever possible)
    List<QuantifiedVariableCore> vars0 = new ArrayList<QuantifiedVariableCore>();
    List<Binder> binders0 = new ArrayList<Binder>();
    Expression exp0 = exp;
    for (int i = vars.size()-1; i >= 0; i--)
    {
      QuantifiedVariableCore var = vars.get(i);
      Binder binder = getBinder(var);
      if (binder == null)
      { 
        vars0.add(var);
        if (!binders0.isEmpty()) 
        {
          Collections.reverse(binders0);
          exp0 = setType(exp, new LetExp(binders0, exp0));
          binders0 = new ArrayList<Binder>();
        }
      }
      else
      {
        binders0.add(binder);
        if (!vars0.isEmpty())
        {
          Collections.reverse(vars0);
          if (forall)
            exp0 = setType(exp, new ForallExp(new QuantifiedVariable(vars0, null), exp0));
          else
            exp0 = setType(exp, new ExistsExp(new QuantifiedVariable(vars0, null), exp0));
          vars0 = new ArrayList<QuantifiedVariableCore>();
        }
      }
    }
    if (!binders0.isEmpty()) 
    {
      Collections.reverse(binders0);
      exp0 = setType(exp, new LetExp(binders0, exp0));
    }
    else if (!vars0.isEmpty())
    {
      Collections.reverse(vars0);
      if (forall)
        exp0 = setType(exp, new ForallExp(new QuantifiedVariable(vars0, null), exp0));
      else
        exp0 = setType(exp, new ExistsExp(new QuantifiedVariable(vars0, null), exp0));
    }
    return exp0;
  }
  
  /***************************************************************************
   * Construct from variable "x:T with x=E" the binder "x=E"
   * @param var the variable (which may or may not be of above form)
   * @return the binder (null, if variable is not of above form)
   **************************************************************************/
  private static Binder getBinder(QuantifiedVariableCore var)
  {
    Expression wexp = var.wexp;
    if (!(wexp instanceof EqualsExp)) return null;
    EqualsExp wexp0 = (EqualsExp)wexp;
    Expression lexp = wexp0.exp1;
    if (!(lexp instanceof IdentifierExp)) return null;
    IdentifierExp lexp0 = (IdentifierExp)lexp;
    if (var.ident.getSymbol() != lexp0.ident.getSymbol()) return null;
    return new Binder(var.ident, wexp0.exp2);
  }
  
  /****************************************************************************
   * Process quantified expression
   * @param qvar the quantified variable.
   * @param the body expression (may be null indicating a reference to the variables)
   * @return the replaced expression (may be the reference to the variables)
   ***************************************************************************/
  private Expression quantifiedVariables(QuantifiedVariable qvar, Expression exp)
  {
    List<Identifier> idents = new ArrayList<Identifier>();
    List<Identifier> idents0 = new ArrayList<Identifier>();
    List<QuantifiedVariableCore> core0 = new ArrayList<QuantifiedVariableCore>();
    int n = qvar.qvcore.length;
    for (int i = 0; i < n; i++)
    {
      QuantifiedVariableCore c = qvar.qvcore[i];
      Expression wexp = c.wexp;
      if (i == n-1 && qvar.exp != null)
      {
        wexp = AST.and(wexp, qvar.exp);
        wexp.setType(qvar.exp.getType());
      }
      Identifier ident = c.ident;
      Identifier ident0 = 
          visible.contains(ident.string) ? freshIdent(ident) : ident;
      idents.add(ident);
      idents0.add(ident0);
      QuantifiedVariableCore c0 = quantifiedVariable(c, wexp, ident0, idents, idents0);
      vars.add(c0);
      core0.add(c0);
    }
    if (exp != null) 
    {
      exp = CloneIds.bind(idents, idExps(idents0), exp);
      return optimizeExpression(exp, polarity);
    }
    if (core0.size() == 1) return identifierExp(core0.get(0).ident);
    List<Expression> exps = new ArrayList<Expression>();
    for (QuantifiedVariableCore c0 : core0) exps.add(identifierExp(c0.ident));
    return new TupleExp(exps);    
  }
  
  /****************************************************************************
   * Get a fresh version of identifier that is currently not visible.
   * @param ident the identifier.
   * @return a derived identifier that is not visible.
   ***************************************************************************/
  private Identifier freshIdent(Identifier ident)
  {
    String string = freshVariable(ident.string);
    Identifier ident0 = new Identifier(string);
    ValueSymbol symbol = (ValueSymbol)ident.getSymbol();
    ident0.setSymbol(new ValueSymbol(ident0, symbol.type, ValueSymbol.Kind.value));
    return ident0;
  }
  
  /****************************************************************************
   * Get the name of a variable that does not occur among the visible variables
   * @param the base name of the variable.
   * @return a derived name which is not visible.
   ***************************************************************************/
  private String freshVariable(String name)
  {
    int i = 0;
    while (true)
    {
      String name0 = "_" +  name + "_" + i;
      if (!visible.contains(name0))  return name0;
      i++;
    }
  }
  
  /****************************************************************************
   * Rename quantified variable core.
   * @param c the original variable core
   * @param wexp the condition attached to the core (may be null)
   * @param ident the new identifier for the core
   * @param core  the original identifiers so far.
   * @param core0 the replacements for the original identifiers so far.
   * @return the renamed version of the core.
   ***************************************************************************/
  private QuantifiedVariableCore quantifiedVariable(
    QuantifiedVariableCore c, Expression wexp, Identifier ident,
    List<Identifier> idents, List<Identifier> idents0)
  { 
    if (c instanceof QuantifiedVariableCore.IdentifierTypeQuantifiedVar)
    {
      QuantifiedVariableCore.IdentifierTypeQuantifiedVar c0 =
          (QuantifiedVariableCore.IdentifierTypeQuantifiedVar)c;
      visible.add(ident.string);
      if (wexp != null)
      { 
        wexp = optimizeFormula(wexp, polarity);
        wexp = CloneIds.bind(idents, idExps(idents0), wexp);
      }
      return new QuantifiedVariableCore.IdentifierTypeQuantifiedVar(
          ident, c0.type, wexp);
    }
    else if (c instanceof QuantifiedVariableCore.IdentifierSetQuantifiedVar)
    {
      QuantifiedVariableCore.IdentifierSetQuantifiedVar c0 =
          (QuantifiedVariableCore.IdentifierSetQuantifiedVar)c;
      Expression exp0 = optimizeTerm(c0.exp);
      exp0.setType(c0.exp.getType());
      visible.add(ident.string);
      if (wexp != null)
      { 
        wexp = optimizeFormula(wexp, polarity);
        wexp = CloneIds.bind(idents, idExps(idents0), wexp);
      }
      return new QuantifiedVariableCore.IdentifierSetQuantifiedVar(
          ident, exp0, wexp);
    }
    return null;
  }
  
  /***************************************************************************
   * Transform list of identifiers into corresponding expressions.
   * @param ids the identifiers.
   * @return the corresponding expressions.
   ***************************************************************************/
  private static List<Expression> idExps(List<Identifier> ids)
  {
    List<Expression> result = new ArrayList<Expression>();
    for (Identifier id : ids)
    {
      Expression exp = new IdentifierExp(id);
      ValueSymbol symbol = (ValueSymbol)id.getSymbol();
      exp.setType(symbol.type);
      result.add(exp);
    }
    return result;
  }
  
  /***************************************************************************
   * Return result expression with type annotated from original expression.
   * @param exp the original expression.
   * @param result the result expression.
   * @return the annotated result expression.
   **************************************************************************/
  private static Expression setType(Expression exp, Expression result)
  {
    result.setType(exp.getType());
    return result;
  }
}
// ---------------------------------------------------------------------------
// end of file
// ---------------------------------------------------------------------------