// ---------------------------------------------------------------------------
// TreeWindow.java
// A Window visualizing an execution trace
// $Id: TreeWindow.java,v 1.19 2022/07/07 10:08:54 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.jku.at>
// Copyright (C) 2018-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package riscal.zest;

import org.eclipse.swt.*;
import org.eclipse.swt.layout.*;
import org.eclipse.swt.widgets.*;
import org.eclipse.swt.graphics.*;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.core.commands.ExecutionException;

import com.google.inject.Guice;
import com.google.inject.Injector;
import com.google.inject.Provider;
import com.google.inject.multibindings.MapBinder;

import javafx.application.*;
import javafx.scene.Scene;
import javafx.scene.input.*;

import java.util.*;
import java.util.List;

import org.eclipse.gef.common.adapt.*;
import org.eclipse.gef.common.adapt.inject.AdapterInjectionSupport;
import org.eclipse.gef.common.adapt.inject.AdapterInjectionSupport.LoggingMode;
import org.eclipse.gef.fx.swt.canvas.*;
import org.eclipse.gef.graph.*;
import org.eclipse.gef.layout.algorithms.*;
import org.eclipse.gef.mvc.fx.domain.*;
import org.eclipse.gef.mvc.fx.viewer.*;
import org.eclipse.gef.mvc.fx.handlers.*;
import org.eclipse.gef.mvc.fx.policies.*;
import org.eclipse.gef.mvc.fx.parts.*;
import org.eclipse.gef.mvc.fx.providers.*;
import org.eclipse.gef.zest.fx.*;
import org.eclipse.gef.zest.fx.parts.*;
import org.eclipse.gef.zest.fx.operations.*;
import org.eclipse.gef.zest.fx.behaviors.*;
import org.eclipse.gef.zest.fx.policies.*;
import org.eclipse.gef.zest.fx.handlers.*;
import org.eclipse.gef.zest.fx.providers.*;
import org.eclipse.gef.mvc.fx.behaviors.*;

import riscal.syntax.*;
import riscal.semantics.*;

public class TreeWindow
{
  // there may be only one instance of this object
  private static TreeWindow TreeWindow;
  
  // the shell of the window
  private Shell shell;

  // the canvas to display the graph on
  private FXCanvasEx canvas;

  // its size
  private int width;
  private int height;
  
  // the map of (nested) graphs to header strings
  private Map<Graph,String> headers = new HashMap<Graph,String>();

  // level of currently displayed procedure invocation
  private int level = 0;
  
  // the root of the tree
  private FormulaTree.Node.Application root;
  
  /****************************************************************************
   * Construct window connected to denoted GUI.
   * @param display the display.
   * @param image the icon image.
   * @param header the header text of the window.
   * @param width the width of the shell
   * @param height the height of the shell
   * @return a window connected to the GUI.
   ***************************************************************************/
  public static TreeWindow construct(Display display, 
    Image image, String header, int width , int height)
  {
    header = "Level 0: " + header;
    Shell shell = new Shell(display, 
        // SWT.SYSTEM_MODAL |
        SWT.CLOSE | SWT.MIN | SWT.MAX | SWT.RESIZE | SWT.TITLE | SWT.BORDER);
    shell.setText(header);
    shell.setImage(image);
    TreeWindow = new TreeWindow(shell, header, width, height);
    return TreeWindow;
  }

  /****************************************************************************
   * Construct panel linked to shell.
   * @param shell the shell
   * @param header the header for the top-level procedure
   * @param width the width of the shell
   * @param height the height of the shell
   ***************************************************************************/
  private TreeWindow(Shell shell, String header, int width, int height)
  {
    // remember shell and its size
    this.shell = shell;
    this.width = width;
    this.height = height;
    
    // set shell layout for shell and panel
    shell.setLayout(new FillLayout());

    // create the graph
    root = FormulaTree.extract();
    if (root == null) return;
    Graph graph = createGraph(header, root);
    if (graph.getNodes().isEmpty()) return;
    
    // visualize the graph
    Platform.setImplicitExit(false);
    canvas = new FXCanvasEx(shell, SWT.NONE);
    Scene scene = createScene(graph);
    canvas.setScene(scene);
    
    // enforce layout on big canvas
    shell.setSize(width, height);
    
    // apparently (?) no way to tell when layout is completed and we may
    // safely resize the window (if done prematurely, layout is broken)
    // try { Thread.sleep(5000); } catch(InterruptedException e) { }
    // shell.setSize(width, height);
  }

  /****************************************************************************
   * Get shell of this trace window.
   ***************************************************************************/
  public Shell getShell() { return shell; }

  /****************************************************************************
   * Create the scene containing the denoted graph.
   * @param graph the graph to be displayed.
   * @return the scene.
   ***************************************************************************/
  private static Scene createScene(Graph graph)
  {    
    Injector injector = Guice.createInjector(new TraceModule());
    IDomain domain = injector.getInstance(IDomain.class);
    IViewer viewer = domain.getAdapter(
        AdapterKey.get(IViewer.class, IDomain.CONTENT_VIEWER_ROLE));  
    Scene scene = new Scene(((IViewer) viewer).getCanvas());
    domain.activate();
    Platform.runLater(new Runnable() 
    {
      @Override
      public void run() 
      {
        viewer.getContents().setAll(Collections.singletonList(graph));
      }
    });
    return scene;
  }

  // -------------------------------------------------------------------------
  //
  // graph construction
  //
  // -------------------------------------------------------------------------

  /***************************************************************************
   * Create a node of the graph.
   * @param label the label of the graph (may be null).
   * @param tooltip the tooltip (may be null).
   * @return the node
   **************************************************************************/
  private static Node node(String label, String tooltip) 
  {
    Node.Builder builder = new Node.Builder();
    if (label != null) builder.attr(ZestProperties.LABEL__NE, label);
    if (tooltip != null) builder.attr(ZestProperties.TOOLTIP__N, tooltip);
    // builder.attr(ZestProperties.LABEL_CSS_STYLE__NE, "-fx-font-family: monospace;");
    return builder.buildNode();
  }

  /***************************************************************************
   * Create an edge of the graph.
   * @param n the source node.
   * @param m the destination node.
   * @param label the label (may be null).
   * @return the edge.
   **************************************************************************/
  private static Edge edge(Node n, Node m, String label) 
  {
    if (label == null) label = "(empty)";
    Edge.Builder builder = new Edge.Builder(n, m);
    if (label != null) builder.attr(ZestProperties.TARGET_LABEL__E, label);
    builder.attr(ZestProperties.TARGET_LABEL_CSS_STYLE__E, 
        "-fx-font-style: italic; -fx-font-weight: bold;");
    builder.attr(ZestProperties.TARGET_DECORATION__E, 
        new javafx.scene.shape.Polygon(0, 0, 5, 2, 5, -2));
    builder.attr(ZestProperties.TARGET_DECORATION_CSS_STYLE__E, 
        "-fx-fill: white;");
    return builder.buildEdge();
  }

  /***************************************************************************
   * Create an dashed edge of the graph.
   * @param n the source node.
   * @param m the destination node.
   * @param label the label (may be null).
   * @return the edge.
   **************************************************************************/
  private static Edge dashed(Node n, Node m, String label) 
  {
    if (label == null) label = "(empty)";
    Edge.Builder builder = new Edge.Builder(n, m);
    if (label != null) builder.attr(ZestProperties.TARGET_LABEL__E, label);
    builder.attr(ZestProperties.TARGET_LABEL_CSS_STYLE__E, 
        "-fx-font-style: italic; -fx-font-weight: bold;");
    builder.attr(ZestProperties.TARGET_DECORATION__E, 
        new javafx.scene.shape.Polygon(0, 0, 5, 2, 5, -2));
    builder.attr(ZestProperties.TARGET_DECORATION_CSS_STYLE__E, 
        "-fx-fill: white;");
    builder.attr(ZestProperties.CURVE_CSS_STYLE__E, 
        "-fx-stroke-dash-array: 3 5;");
    return builder.buildEdge();
  }

 /***************************************************************************
   * Create the graph visualizing the trace
   * @param  header the window header to be displayed for the graph
   * @return the graph.
   **************************************************************************/
  private Graph createGraph(String header, FormulaTree.Node.Application tree)
  { 
    // the nodes and edges generated from the trace
    List<Node> nodes = new ArrayList<Node>();
    List<Edge> edges = new ArrayList<Edge>();

    // populate the graph
    String message;
    Types.Value result = tree.getResult();
    if (result == null)
      message = "";
    else if (tree == root)
      message = "error: " + !((Types.Value.Bool)result).getValue();
    else
      message = "value: " + result;
    String tooltip = append(tree.getPhrase(), message);
    Node node = node(" ", tooltip);
    nodes.add(node);
    level++;
    for (FormulaTree.Node child: tree.getChildren())
      populate(child, node, nodes, edges);
    level--;
    
    HashMap<String, Object> attrs = new HashMap<>();
    if (tree == root)
    {
      SpaceTreeLayoutAlgorithm algorithm;
      algorithm = new SpaceTreeLayoutAlgorithm();
      algorithm.setBranchGap(30);
      algorithm.setLayerGap(30);
      algorithm.setLeafGap(40);
      attrs.put(ZestProperties.LAYOUT_ALGORITHM__G, algorithm);  
    }
    else
    {
      // space tree does not seem to work for nested graphs ???
      attrs.put(ZestProperties.LAYOUT_ALGORITHM__G, 
          new SugiyamaLayoutAlgorithm(SugiyamaLayoutAlgorithm.Direction.VERTICAL,
              new org.eclipse.gef.geometry.planar.Dimension(width, height)));
    }
    Graph graph = new Graph(attrs, nodes, edges);
    headers.put(graph, header);
    return graph;
  }

  /***************************************************************************
   * Append two strings separated by new line.
   * @param s1 the first string.
   * @param s2 the second string.
   * @return the result of the concatenation.
   **************************************************************************/
  private static String append(String s1, String s2)
  {
    if (s1.isEmpty()) return s2;
    if (s1.endsWith("\n")) return s1+s2;
    return s1 + "\n" + s2;
  }
  
  /***************************************************************************
   * Append textual representation of phrase and string.
   * @param phrase the phrase (may be null).
   * @param string the string.
   * @return the result of the concatenation.
   **************************************************************************/
  private static String append(AST phrase, String string)
  {
    if (phrase == null) return string;
    String text = phrase.toString();
    if (text.length() > 40) text = text.substring(0, 40) + "...";
    return append(text, string);
  }
  
  /***************************************************************************
   * Append textual representation of context and string.
   * @param context the context (may be null).
   * @param string the string.
   * @return the result of the concatenation.
   **************************************************************************/
  private static String append(Types.Context context, String string)
  {
    if (context == null) return string;
    return append(context.toString(), string);
  }
  
  /***************************************************************************
   * Return string representation of node value 
   * @param the node (which may have no value)
   * @return the string
   **************************************************************************/
  private static String value(FormulaTree.Node node)
  {
    if (node.hasValue()) return Boolean.toString(node.getValue());
    return "(none)";
  }
  
  /***************************************************************************
   * Return edge label leading to node
   * @param node the target of the edge
   * @return the string
   **************************************************************************/
  private static String edgeLabel(FormulaTree.Node node)
  {
    if (!node.hasValue()) return Integer.toString(node.cnumber);
    return node.cnumber + ":" + node.getValue();
  }
  
  /***************************************************************************
   * Return edge label leading to node
   * @param node the target of the edge
   * @return the string
   **************************************************************************/
  private static String edgeLabel(FormulaTree.Node.Application node)
  {
    Types.Value result = node.getResult();
    if (result == null) return Integer.toString(node.cnumber);
    return node.cnumber + ":" + node.getResult();
  }
  
  /***************************************************************************
   * Populate the graph for the current tree.
   * @param tree the tree from which to construct the graph
   * @param parent the parent node to which to connect the tree
   * @param nodes the list to add new nodes to
   * @param edges the list to add new edges to
   **************************************************************************/
  private void populate(FormulaTree.Node tree, Node parent,
    List<Node> nodes, List<Edge> edges)
  {    
    Node node = null;
    Edge edge = null;
    if (tree instanceof FormulaTree.Node.Arguments)
    {
      FormulaTree.Node.Arguments tree0 = (FormulaTree.Node.Arguments)tree;
      node = node(Integer.toString(tree.cnumber), 
          append(tree.getPhrase(), 
              append(tree.context, "value: " + tree0.getResult())));
      edge = dashed(parent, node, tree.cnumber + ":" + tree0.getResult());
    }
    else if (tree instanceof FormulaTree.Node.Parent)
    {
      node = node(" ", append(tree.getPhrase(), append(tree.context,
          "value: " + value(tree))));
      edge = edge(parent, node, edgeLabel(tree));
    }
    else if (tree instanceof FormulaTree.Node.Application)
    {
      FormulaTree.Node.Application tree0 = (FormulaTree.Node.Application)tree;
      String label = tree.label;
      if (label.length()>10) label = label.substring(0, 8) + "..";
      node = node(label, append(tree.getPhrase(), append(tree.context,
          "value: " + tree0.getResult())));
      edge = dashed(parent, node, edgeLabel(tree0));
      Graph graph = createGraph("Level " + level + ": " + tree.label, tree0);
      if (!graph.getNodes().isEmpty())
        node.setNestedGraph(graph);
    }
    else
    {
      node = node(tree.label, append(tree.getPhrase(), append(tree.context,
          "value: " + value(tree))));
      edge = edge(parent, node, edgeLabel(tree));
    }
    nodes.add(node);
    edges.add(edge);
    if (tree instanceof FormulaTree.Node.Application) return;
    for (FormulaTree.Node child : tree.getChildren())
      populate(child, node, nodes, edges);
  }
  
  // -------------------------------------------------------------------------
  //
  // event handling
  //
  // -------------------------------------------------------------------------

  // the module defining the event handlers
  public static class TraceModule extends ZestFxModule 
  {
    @Override
    protected void enableAdapterMapInjection() 
    {
      install(new AdapterInjectionSupport(LoggingMode.PRODUCTION));
    }
    @Override
    protected void bindNodePartAdapters(MapBinder<AdapterKey<?>, Object> adapterMapBinder) {
      // layout
      adapterMapBinder.addBinding(AdapterKey.defaultRole()).to(NodeLayoutBehavior.class);
      // pruning
      adapterMapBinder.addBinding(AdapterKey.defaultRole()).to(HidePolicy.class);
      adapterMapBinder.addBinding(AdapterKey.defaultRole()).to(ShowHiddenNeighborsPolicy.class);
      adapterMapBinder.addBinding(AdapterKey.defaultRole()).to(NodeHidingBehavior.class);

      // translate on-drag
      adapterMapBinder.addBinding(AdapterKey.defaultRole()).to(TranslateSelectedAndRelocateLabelsOnDragHandler.class);

      // show hidden neighbors on-type
      adapterMapBinder.addBinding(AdapterKey.role("show-hidden-neighbors"))
      .to(ShowHiddenNeighborsOnTypeHandler.class);

      // hide on-type
      adapterMapBinder.addBinding(AdapterKey.role("hide")).to(HideOnTypeHandler.class);

      adapterMapBinder.addBinding(AdapterKey.role("open-nested-graph")).to(TraceOpenNestedGraphOnDoubleClickHandler.class);

      // transform policy for relocation
      adapterMapBinder.addBinding(AdapterKey.defaultRole()).to(TransformPolicy.class);

      // resize policy to resize nesting nodes
      adapterMapBinder.addBinding(AdapterKey.defaultRole()).to(ResizePolicy.class);

      // anchor provider
      adapterMapBinder.addBinding(AdapterKey.defaultRole()).to(NodePartAnchorProvider.class);

      // feedback and handles
      adapterMapBinder
      .addBinding(AdapterKey.role(DefaultSelectionHandlePartFactory.SELECTION_HANDLES_GEOMETRY_PROVIDER))
      .toProvider(new Provider<ShapeBoundsProvider>() {
        @Override
        public ShapeBoundsProvider get() {
          return new ShapeBoundsProvider(1);
        }
      });

      // selection feedback provider
      adapterMapBinder
      .addBinding(AdapterKey.role(DefaultSelectionFeedbackPartFactory.SELECTION_FEEDBACK_GEOMETRY_PROVIDER))
      .toProvider(new Provider<ShapeBoundsProvider>() {
        @Override
        public ShapeBoundsProvider get() {
          return new ShapeBoundsProvider(0.5);
        }
      });

      // selection link feedback provider
      adapterMapBinder
      .addBinding(
          AdapterKey.role(DefaultSelectionFeedbackPartFactory.SELECTION_LINK_FEEDBACK_GEOMETRY_PROVIDER))
      .toProvider(new Provider<ShapeBoundsProvider>() {
        @Override
        public ShapeBoundsProvider get() {
          return new ShapeBoundsProvider(0.5);
        }
      });

      // hover feedback provider
      adapterMapBinder.addBinding(AdapterKey.role(DefaultHoverFeedbackPartFactory.HOVER_FEEDBACK_GEOMETRY_PROVIDER))
      .to(ShapeBoundsProvider.class);

      // focus feedback provider
      adapterMapBinder.addBinding(AdapterKey.role(DefaultFocusFeedbackPartFactory.FOCUS_FEEDBACK_GEOMETRY_PROVIDER))
      .toProvider(new Provider<ShapeBoundsProvider>() {
        @Override
        public ShapeBoundsProvider get() {
          return new ShapeBoundsProvider(0.5);
        }
      });

      // hover on-hover
      adapterMapBinder.addBinding(AdapterKey.defaultRole()).to(HoverOnHoverHandler.class);

      // normalize on drag
      adapterMapBinder.addBinding(AdapterKey.defaultRole()).to(ConnectedSupport.class);
    }

    @Override
    protected void bindEdgePartAdapters(MapBinder<AdapterKey<?>, Object> adapterMapBinder) {
      // layout behavior
      adapterMapBinder.addBinding(AdapterKey.defaultRole()).to(EdgeLayoutBehavior.class);
      // hiding behavior
      adapterMapBinder.addBinding(AdapterKey.defaultRole()).to(EdgeHidingBehavior.class);

      // selection link feedback provider
      adapterMapBinder
          .addBinding(
              AdapterKey.role(DefaultSelectionFeedbackPartFactory.SELECTION_LINK_FEEDBACK_GEOMETRY_PROVIDER))
          .to(GeometricOutlineProvider.class);

      // selection feedback provider
      adapterMapBinder
          .addBinding(AdapterKey.role(DefaultSelectionFeedbackPartFactory.SELECTION_FEEDBACK_GEOMETRY_PROVIDER))
          .to(GeometricOutlineProvider.class);

      // selection handles
      adapterMapBinder
          .addBinding(AdapterKey.role(DefaultSelectionHandlePartFactory.SELECTION_HANDLES_GEOMETRY_PROVIDER))
          .to(GeometricOutlineProvider.class);

      // hover feedback provider
      adapterMapBinder.addBinding(AdapterKey.role(DefaultHoverFeedbackPartFactory.HOVER_FEEDBACK_GEOMETRY_PROVIDER))
          .to(GeometricOutlineProvider.class);

      // geometry provider for focus feedback
      adapterMapBinder.addBinding(AdapterKey.role(DefaultFocusFeedbackPartFactory.FOCUS_FEEDBACK_GEOMETRY_PROVIDER))
          .to(GeometricOutlineProvider.class);

      // clickable area behavior
      adapterMapBinder.addBinding(AdapterKey.defaultRole()).to(ConnectionClickableAreaBehavior.class);

      // translate selected on-drag policy
      adapterMapBinder.addBinding(AdapterKey.defaultRole()).to(TranslateSelectedAndRelocateLabelsOnDragHandler.class);

      // hover on-hover policy
      adapterMapBinder.addBinding(AdapterKey.defaultRole()).to(HoverOnHoverHandler.class);

      // comment this out because otherwise we get exception "Not enough bend points supplied!"
      // adapterMapBinder.addBinding(AdapterKey.defaultRole()).to(BendConnectionPolicy.class);
    }
    
    @Override
    protected void bindIRootPartAdaptersForContentViewer(MapBinder<AdapterKey<?>, Object> adapterMapBinder) {
      super.bindIRootPartAdaptersForContentViewer(adapterMapBinder);

      adapterMapBinder.addBinding(AdapterKey.role("open-parent-graph")).to(TraceOpenParentGraphOnDoubleClickHandler.class);

      // the following are commented out because they trigger error messages

      // keyboard focus traversal
      // adapterMapBinder.addBinding(AdapterKey.defaultRole()).to(TraverseFocusOnTypeHandler.class);

      // select focused on type
      // adapterMapBinder.addBinding(AdapterKey.defaultRole()).to(SelectFocusedOnTypeHandler.class);
    }
  }

  // derived from org.eclipse.gef.zest.fx.handlers.OpenNestedGraphOnDoubleClickHandler
  public static class TraceOpenNestedGraphOnDoubleClickHandler extends AbstractHandler implements IOnClickHandler {

    @Override
    public void click(MouseEvent event) {
      if (event.getClickCount() == 2) {
        // double click, so open nested graph, if it exists
        final Graph nestedGraph = getHost().getContent().getNestedGraph();
        if (nestedGraph != null) {
          IViewer viewer = getHost().getRoot().getViewer();
          try {
            // navigate to nested graph
            viewer.getDomain().execute(new NavigateOperation(viewer, nestedGraph, true),
                new NullProgressMonitor());
            String header = TreeWindow.headers.get(nestedGraph);
            if (header == null) header = "(unknown)";
            TreeWindow.shell.setText(header);
          } catch (ExecutionException e) {
            throw new RuntimeException(e);
          }
        }
      }
    }

    @Override
    public NodePart getHost() {
      return (NodePart) super.getHost();
    }
  }

  // derived from org.eclipse.gef.zest.fx.handlers.OpenParentGraphOnDoubleClickHandler
  public static class TraceOpenParentGraphOnDoubleClickHandler extends AbstractHandler implements IOnClickHandler {

    @Override
    public void click(MouseEvent event) {
      if (event.getClickCount() == 2) {
        // do nothing in case there is an explicit event target
        if (isRegistered(event.getTarget()) && !isRegisteredForHost(event.getTarget())) {
          return;
        }

        // double click, so open nesting graph, if it exists
        IViewer viewer = getHost().getRoot().getViewer();
        final Graph currentGraph = (Graph) viewer.getContents().get(0);
        // due to some strange bug the currentGraph has to be used instead ???
        // final Graph nestingGraph = currentGraph.getNestingNode() != null ? currentGraph.getNestingNode().getGraph()
        //     : null;
        if (currentGraph != null) 
        {
          try {
            // navigate to parent graph
            viewer.getDomain().execute(new NavigateOperation(viewer, currentGraph, false),
                new NullProgressMonitor());
            String header = TreeWindow.headers.get(currentGraph);
            if (header == null) header = "(unknown)";
            TreeWindow.shell.setText(header);
          } catch (ExecutionException e) {
            throw new RuntimeException(e);
          }
        }
      }
    }

    @Override
    public ZestFxRootPart getHost() {
      return (ZestFxRootPart) super.getHost();
    }
  }
}
